# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from gevent import monkey

from volttron.types import MessageBusStopHandler

monkey.patch_socket()
monkey.patch_ssl()
import subprocess
import argparse
import importlib
import logging
import logging.config
import os
import resource
import stat
import sys
from logging import handlers
from pathlib import Path

import gevent

from volttron.server import aip
from volttron.server import server_argparser as config
from volttron.server.containers import service_repo
from volttron.server.logs import (LogLevelAction, configure_logging, log_to_file)
from volttron.server.server_options import ServerOptions
from volttron.utils import ClientContext as cc, execute_command, is_volttron_running
from volttron.utils import (get_version, store_message_bus_config)
from volttron.utils.persistance import load_create_store

_log = logging.getLogger(os.path.basename(sys.argv[0]) if __name__ == "__main__" else __name__)

# No need for str after python 3.8
VOLTTRON_INSTANCES_PATH = Path("~/.volttron_instances").expanduser().resolve()


def run_server():
    """Start the main volttron process.

    Typically, this function is used from main.py and just uses the argparser's
    Options arguments as inputs.   It also can be called with a dictionary.  In
    that case the dictionaries keys are mapped into a value that acts like the
    args options.
    """
    from volttron.types.blinker_events import volttron_home_set_evnt

    os.environ['VOLTTRON_SERVER'] = "1"
    volttron_home = Path(os.environ.get("VOLTTRON_HOME", "~/.volttron")).expanduser()
    os.environ["VOLTTRON_HOME"] = volttron_home.as_posix()

    if is_volttron_running(volttron_home=volttron_home.as_posix()):
        sys.stderr.write(f"Another VOLTTRON instance is already running using this VOLTTRON home {volttron_home}\n")
        sys.exit(1)

    # Raise events that the volttron_home has been set.
    volttron_home_set_evnt.send(run_server)

    if volttron_home.joinpath("config").exists():
        service_repo.add_instance(ServerOptions, ServerOptions(config_file=volttron_home.joinpath("config")))
    else:
        service_repo.add_instance(ServerOptions, ServerOptions(volttron_home=volttron_home))

    server_options: ServerOptions = service_repo.resolve(ServerOptions)

    parser = build_arg_parser(server_options)

    if server_options.messagebus is None:
        raise ValueError("Message Bus Not Found")

    # Parse and expand options
    args = sys.argv[1:]
    conf = os.path.join(volttron_home, "config")
    if os.path.exists(conf) and "SKIP_VOLTTRON_CONFIG" not in os.environ:
        # command line args get preference over same args in config file
        args = args + ["--config", conf]

    opts = parser.parse_args(args)

    # Handle the fact that we don't use store_true and config that requires
    # inverse.  This is not a switch but a mode of operation, so we change
    # from the string to a boolean value here.
    opts.agent_isolation_mode = opts.agent_isolation_mode != 'False'
    dev_mode = opts.dev_mode
    # Update the server options with the command line parameter options.
    server_options.update(opts)
    server_options.store()

    # create poetry project and poetry lock file in VOLTTRON_HOME
    if dev_mode:
        if not (server_options.poetry_project_path / "pyproject.toml").exists():
            raise ValueError("VOLTTRON is run with --dev but unable to find pyproject.toml is current directory - "
                             f"{server_options.poetry_project_path}")
    else:
        setup_poetry_project(server_options.poetry_project_path)

    start_volttron_process(server_options)


def setup_poetry_project(python_project_path: Path):
    toml = os.path.join(python_project_path, "pyproject.toml")
    lockfile = os.path.join(python_project_path, "lockfile")
    if os.path.isfile(toml):
        # try running poetry command to see if python version in pyproject.toml matches the current activated
        # environments python's version
        # ideally if toml exists poetry.lock file should also exist.
        # Run poetry lock. It will error if there is python version mismatch.
        # If python version is right, create lock file or completes successfully without doing anything
            cmd = ["poetry", "lock", "--directory", python_project_path.as_posix()]
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()  # Use p2's output directly
            if p.returncode != 0:
                # ideally we are expecting only a python version mismatch
                errors = stderr.decode().strip().splitlines()
                if "Current Python version" in errors[0] and "is not allowed by the project" in errors[0]:
                    _log.error(errors[0])
                    _log.error("pyproject.toml already exists but the current python version"
                               f"({sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}) does not "
                               "match the project's required version.\n"
                               f"You can point to a new VOLTTRON_HOME or delete pyproject.toml (and poetry.lock) in the "
                               f"current VOLTTRON_HOME directory and restart VOLTTRON to recreate the project for the "
                               f"current python version")
                else:
                    _log.error(errors)
                sys.exit(1)
            else:
                # all is well with existing poetry project return
                return
    try:
        # No toml file.
        # init a poetry project with the current env's python version, and add all the existing env's libraries
        # to the poetry project using poetry add.
        cmd = [
            "poetry", "init", "--directory",
            python_project_path.as_posix(), "--name", "volttron",
            "--python", f"~{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            "--author", "volttron <volttron@pnnl.gov>", "--quiet"
        ]
        execute_command(cmd)
        cmd = [
            "poetry", "--directory",
            python_project_path.as_posix(), "source", "add", "--priority=supplemental", "test-pypi",
            "https://test.pypi.org/simple/"
        ]
        execute_command(cmd)

        # Now add editable source packages
        # Add these first so that these don't get overwritten by non-editable package's dependency
        # i.e. if pointing to volttron-core dir, add this to poetry first so that volttron-lib-auth that depends on
        # volttron-core from pypi won't overwrite the source package
        pip_cmd = ["pip", "list", "--editable"]
        p = subprocess.Popen(pip_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()  # Use p2's output directly
        if p.returncode != 0:
            raise RuntimeError("Error from listing editable packages:", stderr.decode())
        i = 0
        for line in stdout.decode().splitlines():
            print(line)
            i = i + 1
            if i < 3:
                continue
            poetry_cmd = ["poetry", "add", "--directory", python_project_path.as_posix(), "--editable", line.split()[2]]
            p = subprocess.Popen(poetry_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()  # Use p2's output directly
            if p.returncode != 0:
                raise RuntimeError("Unable to add editable package to $VOLTTRON_HOME/pyproject.toml:", stderr.decode())

        # now do multiple piped commands. add all non-editable packages in one go using piped cmd
        pip_cmd = ["pip", "list", "--format", "freeze",  "--exclude-editable"]
        # Second command
        grep_cmd = ["grep", "-v", "volttron=="]
        # Third command
        poetry_cmd = ["xargs", "poetry", "add", "--directory", python_project_path.as_posix()]

        err_msg = ""
        # Execute the first command
        p1 = subprocess.Popen(pip_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p1_stdout, p1_stderr = p1.communicate()
        # Check and print the output of the first command
        if p1.returncode != 0:
            err_msg = "Error from pip command:", p1_stderr.decode()
        else:
            # Execute the second command, with stdin from the first command's stdout
            p2 = subprocess.Popen(grep_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            p2_stdout, p2_stderr = p2.communicate(input=p1_stdout)    # Use p1's output directly
            # Check and print the output of the second command
            if p2.returncode != 0:
                err_msg = "Error from grep command:", p2_stderr.decode()
            else:
                # Check if `grep` produced any output
                if not p2_stdout.strip():
                    print("No packages to add; grep output is empty.")
                else:
                    # Execute the third command, with stdin from the second command's stdout
                    p3 = subprocess.Popen(poetry_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    p3_stdout, p3_stderr = p3.communicate(input=p2_stdout)    # Use p2's output directly
                    if p3.returncode != 0:
                        err_msg = "Error from poetry command:", p3_stderr.decode()

        if err_msg:
            err_msg = (f"Unable to update pyproject.toml in {python_project_path.as_posix()} with venv's current list of "
                       f"libs. {err_msg}")
            raise RuntimeError(err_msg)
    except RuntimeError:
        # in case of any errors, remove pyproject.toml and poetry.lock if it exists.
        if os.path.exists(toml):
            os.remove(toml)
        if os.path.exists(lockfile):
            os.remove(lockfile)
        raise

def start_volttron_process(options: ServerOptions):
    opts = options
    # Change working dir
    os.chdir(opts.volttron_home)

    if opts.enable_federation:
        if not opts.address or len(opts.address) == 0:
            _log.error("Cannot enable federation: no external address available.")
            _log.error("Use --address to specify an external address for other platforms to connect to.")
            _log.error("Federation will not be enabled.")
            # Disable federation to prevent further attempts
            opts.enable_federation = False
            sys.exit(1)
            

    # vip_address is meant to be a list so make it so.
    if not isinstance(opts.address, list):
        opts.address = [opts.address]
    if opts.log:
        opts.log = config.expandall(opts.log)
    if opts.log_config:
        opts.log_config = config.expandall(opts.log_config)

    # Configure logging
    level = max(1, opts.verboseness)
    if opts.monitor and level > logging.INFO:
        level = logging.INFO

    if opts.log is None:
        log_to_file(sys.stderr, level)
    elif opts.log == "-":
        log_to_file(sys.stdout, level)
    elif opts.log:
        log_to_file(opts.log, level, handler_class=handlers.WatchedFileHandler)
    else:
        log_to_file(None, 100, handler_class=lambda x: logging.NullHandler())

    if opts.log_config:
        with open(opts.log_config, "r") as f:
            for line in f.readlines():
                _log.info(line.rstrip())

        error = configure_logging(opts.log_config)

        if error:
            _log.error("{}: {}".format(*error))
            sys.exit(1)

    if opts.agent_isolation_mode == "True":
        _log.info("VOLTTRON starting in secure mode")
        os.umask(0o007)
    else:
        opts.agent_isolation_mode = "False"

    logging.getLogger("watchdog.observers.inotify_buffer").setLevel(logging.INFO)

    opts.address = [config.expandall(addr) for addr in opts.address]
    opts.local_address = config.expandall(opts.local_address)

    opts.messagebus = config.expandall(opts.messagebus)

    os.environ["MESSAGEBUS"] = opts.messagebus
    os.environ["AGENT_ISOLATION_MODE"] = opts.agent_isolation_mode
    if opts.instance_name is None:
        if len(opts.address) > 0:
            opts.instance_name = opts.address[0]

    _log.debug("opts.instancename {}".format(opts.instance_name))
    if opts.instance_name:
        store_message_bus_config(opts.messagebus, opts.instance_name)
    else:
        # if there is no _instance_name given get_platform_instance_name will
        # try to retrieve from config or default a value and store it in the config
        cc.get_instance_name()

    # Log configuration options
    if getattr(opts, "show_config", False):
        _log.info("volttron version: {}".format(get_version()))
        for name, value in sorted(vars(opts).items()):
            _log.info("%s: %s" % (name, str(repr(value))))

    # Increase open files resource limit to max or 8192 if unlimited
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)

    except OSError:
        _log.exception("error getting open file limits")
    else:
        if soft != hard and soft != resource.RLIM_INFINITY:
            try:
                limit = 8192 if hard == resource.RLIM_INFINITY else hard
                resource.setrlimit(resource.RLIMIT_NOFILE, (limit, hard))
            except OSError:
                _log.exception("error setting open file limits")
            else:
                _log.debug(
                    "open file resource limit increased from %d to %d",
                    soft,
                    limit,
                )
        _log.debug("open file resource limit %d to %d", soft, hard)

    # import all essential volttron types
    from volttron.loader import load_subclasses, find_subpackages

    if opts.auth_enabled:
        # Shouldn't default VolttronAuthService impl of AuthService also be loaded dynamically?
        # If so it should be moved to volttron.auth and not be in volttron.services.auth
        #import volttron.services.auth
        # Load the package
        auth_pkg = importlib.import_module('volttron.types.auth')
        # Access the __all__ attribute
        base_auth_classes = getattr(auth_pkg, '__all__', None)
        # load classes from volttron-lib-auth + any packages defined by additional libraries
        for cls in base_auth_classes:
            load_subclasses('volttron.types.auth.'+cls, "volttron.auth")

    if opts.enable_federation:
        from volttron.services.federation import FederationService

    aip_platform = service_repo.resolve(aip.AIPplatform)
    aip_platform.setup()
    opts.aip = aip_platform

    # Check for secure mode/permissions on VOLTTRON_HOME directory
    mode = os.stat(opts.volttron_home).st_mode
    if mode & (stat.S_IWGRP | stat.S_IWOTH):
        _log.warning("insecure mode on directory: %s", opts.volttron_home)

    external_address_file = os.path.join(opts.volttron_home, "external_address.json")
    _log.debug("external_address_file file %s", external_address_file)

    if opts.agent_monitor_frequency:
        try:
            int(opts.agent_monitor_frequency)
        except ValueError as e:
            raise ValueError("agent-monitor-frequency should be integer "
                             "value. Units - seconds. This determines how "
                             "often the platform checks for any crashed agent "
                             "and attempts to restart. {}".format(e))

    pid_file = os.path.join(opts.volttron_home.as_posix(), "VOLTTRON_PID")
    try:
        _log.debug("********************************************************************")
        _log.debug("VOLTTRON PLATFORM RUNNING ON {} MESSAGEBUS".format(opts.messagebus))
        _log.debug("********************************************************************")
        from volttron.server.decorators import start_service_agents
        from volttron.services.config_store.config_store_service import \
            ConfigStoreService
        from volttron.services.control.control_service import ControlService
        from volttron.services.health.health_service import HealthService

        from volttron.types.known_host import \
            KnownHostProperties as known_host_properties
        from volttron.utils import jsonapi

        spawned_greenlets = []

        # TODO Replace with module level zmq that holds all of the zmq bits in order to start and
        #  run the message bus regardless of whether it's zmq or rmq.

        from volttron.types import MessageBus
        # A message bus is required, if we don't have one installed then this will definitely fail.
        mb: MessageBus = service_repo.resolve(MessageBus) # type: ignore
        auth_service = None
        # TODO move this to laoder code
        if options.auth_enabled:
            from volttron.types.auth import Authenticator
            from volttron.types.auth.auth_service import AuthService
            authenticator = service_repo.resolve(Authenticator)
            auth_service = service_repo.resolve(AuthService)

        # First load auth_service so that config_store can use
        # auth service to wrap protected_rpcs
        if auth_service is not None:
            event = gevent.event.Event()
            task = gevent.spawn(auth_service.core.run, event)
            event.wait()
            del event
            spawned_greenlets.append(task)

        config_store = service_repo.resolve(ConfigStoreService)
        event = gevent.event.Event()
        task = gevent.spawn(config_store.core.run, event)
        event.wait()
        del event
        spawned_greenlets.append(task)

        class StopHandler(MessageBusStopHandler):

            def message_bus_shutdown(self):
                for spawned_task in spawned_greenlets:
                    spawned_task.kill(block=False)

        mb.set_stop_handler(StopHandler())

        # from volttron.server.decorators import get_messagebus_class
        # MessageBusClass = get_messagebus_class()
        # # Allows registration agents to callbacks for peers
        # notifier = ServicePeerNotifier()
        #
        # mb = MessageBusClass(opts, notifier, tracker, protected_topics, external_address_file, config_store.core.stop)

        mb.start()
        elapsed_time = 0
        while elapsed_time < 30:
            if mb.is_running():
                break
            gevent.sleep(1)
            elapsed_time += 1

        # if address is already bound to another instance mb start will fail.
        if not mb.is_running():
            sys.stderr.write("Messagebus did not start within the timeout period. Please check logs\n")
            sys.exit(1)

        # The instance file is where we are going to record the instance and
        # its details according to
        if VOLTTRON_INSTANCES_PATH.is_file():
            instances = jsonapi.loads(VOLTTRON_INSTANCES_PATH.open().read())
        else:
            instances = {}

        # Trim instances from instance file.
        # Need to use list here because we are going to potentially decrease the size of the dictionary so
        # we need to be careful about the iteration.
        for k in list(instances):
            try:
                # Raises OSError if the pid is not found note this will not actually kill the process.
                # https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python/6940314#6940314
                os.kill(instances[k]['pid'], 0)
            except OSError:
                del instances[k]

        this_instance = instances.get(opts.volttron_home.as_posix(), {})
        this_instance["pid"] = os.getpid()
        this_instance["version"] = get_version()

        # note vip_address is a list
        this_instance["address"] = opts.address
        this_instance["volttron-home"] = opts.volttron_home.as_posix()
        this_instance["start-args"] = sys.argv[1:]
        instances[opts.volttron_home.as_posix()] = this_instance

        external_address_file = os.path.join(opts.volttron_home, "external_address.json")
        _log.debug("external_address_file file %s", external_address_file)

        with VOLTTRON_INSTANCES_PATH.open("w") as fp:
            fp.write(jsonapi.dumps(instances))

        # TODO - health service should register callback with messagebus to keep track of peer add and remove
        # health_service = service_repo.resolve(HealthService)
        # event = gevent.event.Event()
        # task = gevent.spawn(health_service.core.run, event)
        # event.wait()
        # del event
        # spawned_greenlets.append(task)
        # mb._notifier.register_peer_callback(health_service.peer_added,
        #                                     health_service.peer_dropped)

        spawned_greenlets.extend(start_service_agents())

        # # # TODO Key discovery agent add in.
        # # # KeyDiscoveryAgent(
        # # #     address=address,
        # # #     serverkey=publickey,
        # # #     identity=KEY_DISCOVERY,
        # # #     external_address_config=external_address_file,
        # # #     setup_mode=opts.setup_mode,
        # # #     bind_web_address=opts.bind_web_address,
        # # #     enable_store=False,
        # # #     message_bus="zmq",
        # # # ),
        # # ]

        # Auto-start agents now that all services are up
        if opts.autostart:
            for name, error in opts.aip.autostart():
                _log.error("error starting {!r}: {}\n".format(name, error))

        # Done with all start up process write a PID file

        with open(pid_file, "w+") as f:
            f.write(str(os.getpid()))

        if opts.enable_federation:
            # Touch the file to cause a reload for the watcher.
            fedfile = opts.volttron_home / "federation_config.json"
            fedfile.touch()

        _log.debug("Finished Startup of Platform.")

        # Wait for any service to stop, signaling exit
        try:
            gevent.wait(spawned_greenlets, count=1)
        except KeyboardInterrupt:
            _log.info("SIGINT received; shutting down")
        finally:
            sys.stderr.write("Shutting down.\n")
            _log.debug("Kill all service agent tasks")
            for task in spawned_greenlets:
                task.kill(block=False)
            gevent.wait(spawned_greenlets)
    except Exception as e:
        _log.error(e)
        import traceback
        _log.error(traceback.print_exc())
    finally:
        _log.debug("AIP finally")
        opts.aip.finish()
        instance_file = str(VOLTTRON_INSTANCES_PATH)
        try:
            instances = load_create_store(instance_file)
            instances.pop(opts.volttron_home, None)
            instances.sync()
            if os.path.exists(pid_file):
                os.remove(pid_file)
        except Exception:
            _log.warning(f"Unable to load {VOLTTRON_INSTANCES_PATH}")
        _log.debug("********************************************************************")
        _log.debug("VOLTTRON PLATFORM HAS SHUTDOWN")
        _log.debug("********************************************************************")


def build_arg_parser(options: ServerOptions) -> argparse.ArgumentParser:
    """
    Builds and returns an argument parser.

    :return: The argument parser.
    :rtype: argparse.ArgumentParser
    """

    from volttron.server.server_options import ServerOptions
    from volttron.types.events import volttron_home_set_evnt

    default_levels_for_modules = {
        "volttron.server.decorators": logging.WARNING,
        "volttron.server.containers": logging.INFO,
        "volttron.loader": logging.WARNING,
        "volttron.server.run_server": logging.INFO,
        "volttron.client.decorators": logging.INFO,
    # "volttron.messagebus.zmq.socket": logging.INFO
    }
    [logging.getLogger(k).setLevel(v) for k, v in default_levels_for_modules.items()]

    volttron_home = os.path.normpath(config.expandall(os.environ.get("VOLTTRON_HOME", "~/.volttron")))
    os.environ["VOLTTRON_HOME"] = volttron_home

    argv = sys.argv

    # Refuse to run as root
    if not getattr(os, "getuid", lambda: -1)():
        sys.stderr.write("%s: error: refusing to run as root to prevent "
                         "potential damage.\n" % os.path.basename(argv[0]))
        sys.exit(77)

    # Setup option parser
    parser = config.ArgumentParser(
        prog=os.path.basename(argv[0]),
        add_help=False,
        description="VOLTTRON platform service",
        usage="%(prog)s [OPTION]...",
        argument_default=argparse.SUPPRESS,
        epilog="Boolean options, which take no argument, may be inversed by "
        "prefixing the option with no- (e.g. --autostart may be "
        "inversed using --no-autostart).",
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="FILE",
        action="parse_config",
        ignore_unknown=False,
        sections=[None, "volttron"],
        help="read configuration from FILE",
    )
    parser.add_argument(
        "-l",
        "--log",
        metavar="FILE",
        default=None,
        help="send log output to FILE instead of stderr",
    )
    parser.add_argument(
        "-L",
        "--log-config",
        metavar="FILE",
        help="read logging configuration from FILE",
    )
    parser.add_argument(
        "--log-level",
        metavar="LOGGER:LEVEL",
        action=LogLevelAction,
        help="override default logger logging level",
    )
    parser.add_argument(
        "--monitor",
        action="store_true",
        help="monitor and log connections (implies -v)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="add_const",
        const=10,
        dest="verboseness",
        help="decrease logger verboseness; may be used multiple times",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="add_const",
        const=-10,
        dest="verboseness",
        help="increase logger verboseness; may be used multiple times",
    )
    parser.add_argument(
        "--verboseness",
        type=int,
        metavar="LEVEL",
        default=logging.WARNING,
        help="set logger verboseness",
    )
    parser.add_argument("--messagebus",
                        type=str,
                        default=options.messagebus,
                        help="The message bus to use during startup.")
    # parser.add_argument("--auth-service",
    #                     type=str,
    #                     default=options.auth_service,
    #                     help="The auth service to use for authentication of clients.")
    # parser.add_argument("--authentication-class",
    #                     type=str,
    #                     default=options.authentication_class,
    #                     help="Class used with the AuthService for authentication")
    # parser.add_argument("--authorization-class",
    #                     type=str,
    #                     default=options.authorization_class,
    #                     help="Class used with the AuthService for authorization")
    # parser.add_argument(
    #    '--volttron-home', env_var='VOLTTRON_HOME', metavar='PATH',
    #    help='VOLTTRON configuration directory')
    parser.add_argument("--auth-enabled",
                        action="store_true",
                        inverse="--auth-disabled",
                        dest="auth_enabled",
                        help=argparse.SUPPRESS)
    parser.add_argument("--auth-disabled", action="store_false", help=argparse.SUPPRESS)
    parser.add_argument("--show-config", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--dev",
                        action="store_true",
                        dest="dev_mode",
                        default=False,
                        help="development mode with poetry environment to build volttron libraries from source/pypi")

    parser.add_help_argument()
    parser.add_version_argument(version="%(prog)s " + str(get_version()))

    agents = parser.add_argument_group("agent options")
    agents.add_argument(
        "--autostart",
        action="store_true",
        inverse="--no-autostart",
        help="automatically start enabled agents and services",
    )
    agents.add_argument(
        "--no-autostart",
        action="store_false",
        dest="autostart",
        help=argparse.SUPPRESS,
    )

    agents.add_argument(
        "--address",
        metavar="MESSAGE_BUS_ADDR",
        action="append",
        default=[],
        help="Address for binding to the message bus. Required for Federation.",
    )
    agents.add_argument(
        "--local-address",
        metavar="ZMQADDR",
        help="ZeroMQ URL to bind for local agent VIP connections",
    )

    agents.add_argument(
        "--instance-name",
        default=options.instance_name,
        help="The name of the VOLTTRON instance this command is starting up.",
    )
    agents.add_argument(
        "--enable-federation",
        action="store_true",
        inverse="--disable-federation",
        help="Enable platform federation for connecting multiple VOLTTRON instances",
    )
    agents.add_argument(
        "--disable-federation",
        action="store_false",
        dest="enable_federation",
        help=argparse.SUPPRESS,
    )
    agents.add_argument(
        "--federation-url",
        default=None,
        help="URL of federation registry service to connect to. Requires --address to be specified.",
    )
    # TODO: Determine if we need this
    # agents.add_argument(
    #     "--msgdebug",
    #     action="store_true",
    #     help="Route all messages to an agent while debugging.",
    # )
    # agents.add_argument(
    #     "--setup-mode",
    #     action="store_true",
    #     help="Setup mode flag for setting up authorization of external platforms.",
    # )
    agents.add_argument(
        "--agent-monitor-frequency",
        default=600,
        help="How often should the platform check for crashed agents and "
        "attempt to restart. Units=seconds. Default=600",
    )
    agents.add_argument(
        "--agent-isolation-mode",
        default=False,
        help="Require that agents run with their own users (this requires "
        "running scripts/secure_user_permissions.sh as sudo)",
    )

    agents.add_argument("--server-messagebus-id", default="vip.server", help="A connection from the server")

    ipc = "ipc://%s$VOLTTRON_HOME/run/" % ("@" if sys.platform.startswith("linux") else "")

    parser.set_defaults(log=None,
                        log_config=None,
                        monitor=False,
                        verboseness=logging.WARNING,
                        volttron_home=options.volttron_home,
                        autostart=True,
                        address=options.address,
                        local_address=ipc + "vip.socket",
                        instance_name=options.instance_name,
                        resource_monitor=True,
                        msgdebug=None,
                        setup_mode=False,
                        agent_isolation_mode=False,
                        server_messagebus_id="vip.server",
                        enable_federation=False,
                        federation_url=None)

    return parser


# def main(argv=sys.argv):
#     import coloredlogs

#     from volttron.server.server_options import ServerOptions
#     from volttron.types.events import volttron_home_set_evnt
#     from volttron.utils.logs import setup_logging

#     # Refuse to run as root
#     if not getattr(os, "getuid", lambda: -1)():
#         sys.stderr.write("%s: error: refusing to run as root to prevent "
#                          "potential damage.\n" % os.path.basename(argv[0]))
#         sys.exit(77)

#     default_levels_for_modules = {
#         "volttron.server.decorators": logging.WARNING,
#         "volttron.server.containers": logging.INFO,
#         "volttron.loader": logging.WARNING,
#         "volttron.server.run_server": logging.INFO,
#         "volttron.client.decorators": logging.INFO,
#         "volttron.messagebus.zmq.socket": logging.INFO
#     }
#     [logging.getLogger(k).setLevel(v) for k, v in default_levels_for_modules.items()]

#     volttron_home = os.path.normpath(
#         config.expandall(os.environ.get("VOLTTRON_HOME", "~/.volttron")))
#     os.environ["VOLTTRON_HOME"] = volttron_home

#     #load_volttron_packages()
#     # Setup option parser
#     parser = config.ArgumentParser(
#         prog=os.path.basename(argv[0]),
#         add_help=False,
#         description="VOLTTRON platform service",
#         usage="%(prog)s [OPTION]...",
#         argument_default=argparse.SUPPRESS,
#         epilog="Boolean options, which take no argument, may be inversed by "
#         "prefixing the option with no- (e.g. --autostart may be "
#         "inversed using --no-autostart).",
#     )
#     parser.add_argument(
#         "-c",
#         "--config",
#         metavar="FILE",
#         action="parse_config",
#         ignore_unknown=False,
#         sections=[None, "volttron"],
#         help="read configuration from FILE",
#     )
#     parser.add_argument(
#         "-l",
#         "--log",
#         metavar="FILE",
#         default=None,
#         help="send log output to FILE instead of stderr",
#     )
#     parser.add_argument(
#         "-L",
#         "--log-config",
#         metavar="FILE",
#         help="read logging configuration from FILE",
#     )
#     parser.add_argument(
#         "--log-level",
#         metavar="LOGGER:LEVEL",
#         action=LogLevelAction,
#         help="override default logger logging level",
#     )
#     parser.add_argument(
#         "--monitor",
#         action="store_true",
#         help="monitor and log connections (implies -v)",
#     )
#     parser.add_argument(
#         "-q",
#         "--quiet",
#         action="add_const",
#         const=10,
#         dest="verboseness",
#         help="decrease logger verboseness; may be used multiple times",
#     )
#     parser.add_argument(
#         "-v",
#         "--verbose",
#         action="add_const",
#         const=-10,
#         dest="verboseness",
#         help="increase logger verboseness; may be used multiple times",
#     )
#     parser.add_argument(
#         "--verboseness",
#         type=int,
#         metavar="LEVEL",
#         default=logging.WARNING,
#         help="set logger verboseness",
#     )
#     # parser.add_argument(
#     #    '--volttron-home', env_var='VOLTTRON_HOME', metavar='PATH',
#     #    help='VOLTTRON configuration directory')
#     parser.add_argument("--show-config", action="store_true", help=argparse.SUPPRESS)
#     parser.add_help_argument()
#     parser.add_version_argument(version="%(prog)s " + str(get_version()))

#     agents = parser.add_argument_group("agent options")
#     agents.add_argument(
#         "--autostart",
#         action="store_true",
#         inverse="--no-autostart",
#         help="automatically start enabled agents and services",
#     )
#     agents.add_argument(
#         "--no-autostart",
#         action="store_false",
#         dest="autostart",
#         help=argparse.SUPPRESS,
#     )
#     agents.add_argument(
#         "--address",
#         metavar="ZMQADDR",
#         action="append",
#         default=[],
#         help="ZeroMQ URL to bind for VIP connections",
#     )
#     agents.add_argument(
#         "--local-address",
#         metavar="ZMQADDR",
#         help="ZeroMQ URL to bind for local agent VIP connections",
#     )
#     agents.add_argument(
#         "--instance-name",
#         default=None,
#         help="The name of the instance that will be reported to "
#         "VOLTTRON central.",
#     )
#     agents.add_argument(
#         "--msgdebug",
#         action="store_true",
#         help="Route all messages to an agent while debugging.",
#     )
#     agents.add_argument(
#         "--setup-mode",
#         action="store_true",
#         help="Setup mode flag for setting up authorization of external platforms.",
#     )
#     parser.add_argument(
#         "--messagebus",
#         action="store",
#         default="zmq",
#         dest="messagebus",
#         help="set message to be used. valid values are zmq and rmq",
#     )
#     agents.add_argument(
#         "--agent-monitor-frequency",
#         default=600,
#         help="How often should the platform check for crashed agents and "
#         "attempt to restart. Units=seconds. Default=600",
#     )
#     agents.add_argument(
#         "--agent-isolation-mode",
#         default=False,
#         help="Require that agents run with their own users (this requires "
#         "running scripts/secure_user_permissions.sh as sudo)",
#     )

#     # XXX: re-implement control options
#     # on
#     # control.add_argument(
#     #    '--allow-root', action='store_true', inverse='--no-allow-root',
#     #    help='allow root to connect to control socket')
#     # control.add_argument(
#     #    '--no-allow-root', action='store_false', dest='allow_root',
#     #    help=argparse.SUPPRESS)
#     # control.add_argument(
#     #    '--allow-users', action='store_list', metavar='LIST',
#     #    help='users allowed to connect to control socket')
#     # control.add_argument(
#     #    '--allow-groups', action='store_list', metavar='LIST',
#     #    help='user groups allowed to connect to control socket')

#     ipc = "ipc://%s$VOLTTRON_HOME/run/" % ("@" if sys.platform.startswith("linux") else "")

#     parser.set_defaults(
#         log=None,
#         log_config=None,
#         monitor=False,
#         verboseness=logging.WARNING,
#         volttron_home=volttron_home,
#         autostart=True,
#         address=[],
#         local_address=ipc + "vip.socket",
#         _instance_name=None,
#     # allow_root=False,
#     # allow_users=None,
#     # allow_groups=None,
#         verify_agents=True,
#         resource_monitor=True,
#     # mobility=True,
#         msgdebug=None,
#         setup_mode=False,
#     # Type of underlying message bus to use - ZeroMQ or RabbitMQ
#         messagebus="zmq",
#     )

#     # Parse and expand options
#     args = argv[1:]
#     conf = os.path.join(volttron_home, "config")
#     if os.path.exists(conf) and "SKIP_VOLTTRON_CONFIG" not in os.environ:
#         # command line args get preference over same args in config file
#         args = args + ["--config", conf]
#     logging.getLogger().setLevel(logging.NOTSET)
#     opts = parser.parse_args(args)

#     load_messagebus_module(opts.messagebus)

#     start_volttron_process(opts)


def _main():
    """Entry point for scripts."""
    try:
        sys.exit(run_server())
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    _main()



import inspect
import logging
import typing
from typing import TYPE_CHECKING, Optional, TypeVar

from volttron.client.vip.agent.core import Core
from volttron.types.auth.auth_credentials import Credentials
from volttron.types.factories import (ConnectionBuilder, ControlParser, CoreBuilder)

_log = logging.getLogger(__name__)

T = TypeVar('T')


def factory_registration(registry_name: str, protocol: T = None, singleton: bool = True, allow_many: bool = False):
    """
    Create a factory registration function.

    The function will have a registry attribute that is a dictionary of registered
    classes and a name attribute that is the name of the factory.

    @param name: The name of the factory.
    @type name: str
    @return: The factory registration function.
    @rtype: function
    """

    def register(cls, **kwargs):
        lookup_key = None

        if hasattr(cls, 'Meta'):
            # Meta can either have a name or an identity, but not both.
            # We use either one of the values as a lookup for the register.
            if hasattr(cls.Meta, 'name') and hasattr(cls.Meta, 'identity'):
                raise ValueError("Only name or identity can be specified in Meta.")
            elif hasattr(cls.Meta, 'name'):
                lookup_key = cls.Meta.name
            elif hasattr(cls.Meta, 'identity'):
                lookup_key = cls.Meta.identity
        else:
            lookup_key = cls.__name__

        if lookup_key is None:
            raise ValueError(f"{cls.__name__} does not have an internal Meta class with identity or name.")

        # args = typing.get_args(protocol)
        # if
        if protocol is not None and not protocol in cls.__bases__ and not isinstance(cls, protocol):
            raise ValueError(f"{cls.__name__} doesn't implement {protocol}")

        # if singleton:
        #     if allow_many:
        #         service_repo.add_concrete_reference(protocol, cls, kwargs=kwargs)
        #     else:
        #         service_repo.add_interface_reference(protocol, cls, kwargs=kwargs)
        # else:
        #     service_repo.add_factory(cls, cls, kwargs=kwargs)

        if lookup_key is None:
            _log.warning("Lookup key is none!")
        _log.debug(f"Registering {cls.__name__} as a {lookup_key}")
        if lookup_key in register.registry:
            _log.warning(f"{lookup_key} already in register for {register.registry_name}.")
            #raise ValueError(f"{lookup_key} already in register for {register.registry_name}.")
        register.registry[lookup_key] = cls
        return cls

    register.registry_name = registry_name
    register.registry = {}
    return register


core_builder = factory_registration("core_builder", protocol=CoreBuilder, singleton=True)
connection_builder = factory_registration("connection_bulider", protocol=ConnectionBuilder, singleton=True)
vctl_subparser = factory_registration("vctl_subparser", protocol=ControlParser)


def __get_class_from_factory__(*, registration, name: str = None):
    if not registration.registry:
        raise ValueError(f"No {name} is currently registered")

    if name is None and len(registration.registry) > 1:
        raise ValueError(f"Can't figure out which messagebus to return.")

    if name is None:
        # First name of the registry dictionary.
        name = list(registration.registry.keys())[0]

    if name not in registration.registry:
        raise ValueError(f"Couldn't retrieve {name} from register")

    return registration.registry.get(name)


__core_builder__: CoreBuilder = None


def get_core_builder(name: Optional[str] = None, **kwargs) -> CoreBuilder:
    global __core_builder__
    # TODO: Not sure why the second lookup doesn't work as it should.
    # TODO: Server is going to use a file in volttron_home called system to hold the
    # TODO: Builder so we can get it if we are a client.
    if __core_builder__ is None:
        try:
            __core_builder__ = __get_class_from_factory__(registration=core_builder, name=name)
        except ValueError:

            import importlib
            zmq_core_module = "volttron.messagebus.zmq.zmq_core"
            zmq_core_builder_class = "ZmqCoreBuilder"
            module = importlib.import_module(zmq_core_module)
            __core_builder__ = __get_class_from_factory__(
                registration=core_builder, name=name)    # __core_builder__ = getattr(module, zmq_core_builder_class)

            # __core_builder__ = importlib.import_module(new_package)

        specs = inspect.getfullargspec(__core_builder__.__init__)
        # for k, v in kwargs.items():
        #     if k not in signature.parameters:
        #         raise ValueError(f"Invalid parameter {k} signature has {signature.parameters}")
        if len(specs.args) == 1 and specs.defaults is None:
            __core_builder__ = __core_builder__()
        else:
            __core_builder__ = __core_builder__(**kwargs)

    return __core_builder__


def get_connection_builder(name: Optional[str] = None) -> ConnectionBuilder:
    return __get_class_from_factory__(registration=connection_builder, name=name)


def get_server_credentials(address: Optional[str] = None) -> Credentials:
    import os
    from pathlib import Path

    from volttron.types.auth import (Credentials, PKICredentials, PublicCredentials, VolttronCredentials)
    from volttron.types.known_host import KnownHostProperties as known_host_properties
    from volttron.client.known_identities import PLATFORM
    from volttron.utils import jsonapi

    # ipc address must mean we are local so use @ symbol to mean so.
    if address is None or address.startswith('ipc'):
        address = "@"

    cred_path = Path(os.environ['VOLTTRON_HOME']).expanduser() / f"credentials_store/{PLATFORM}.json"
    return VolttronCredentials.load_from_file(cred_path)

    new_path = Path(os.environ['VOLTTRON_HOME']) / "known_hosts.json"
    if known_host_properties is None:
        from volttron.types.known_host import _KnownHostProperties
        known_host_properties = _KnownHostProperties.load(new_path)

        # Handle original known_hosts file if necessary.
        old_path = Path(os.environ['VOLTTRON_HOME']) / "known_hosts"
        if old_path.exists():
            old_data = jsonapi.loads(old_path.open().read())
            if "@" in old_data:
                known_host_properties.add_property("@", "publickey", old_data["@"])

        known_host_properties.store(new_path)

    data = jsonapi.loads((Path(os.environ["VOLTTRON_HOME"]) / "keystore").open().read())

    # TODO This should only be necessary if there is in fact a credential store.
    # TODO Redo this thing here!
    known_host_properties.add_property(address, "publickey", data["public"])
    known_host_properties.store(new_path)
    publickey = known_host_properties.get_property(address, "publickey")

    # TODO This should only return the PublicCredentials instead of PKICredentials, but for now
    # TODO Verify the credentails for the server.
    return PKICredentials(identity="platform", publickey=data['public'], secretkey=data['secret'])

    # credential_type = os.environ.get("VOLTTRON_CREDENTIALS_TYPE")
    # credentials = os.environ.get("VOLTTRON_SERVER_CREDENTIALS")

    # if credentials is None:
    #     return Credentials.create(identity="server")

    # if credentials is not None and credential_type is None:
    #     raise ValueError("env VOLTTRON_CREDENTIALS is set but VOLTTRON_CREDENTIAL_TYPE is not set.")

    # if credentials is not None and credential_type is not None:
    #     if credential_type == "public":
    #         return PublicCredentials.from_json(Path(credentials).open().read())
    #     elif credential_type == "pki":
    #         return PKICredentials.from_json(Path(credentials).open().read())
    #     else:
    #         raise ValueError(f"Invalid VOLTTRON_CREDENTIAL_TYPE: {credential_type}")



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
"""
Advanced argument parser.

Fully compatible with argparse, and can be used as a drop-in
replacement, with the added ability add options from configuration files
and environment variables. The default order of precedense is
command-line > environment > config file > defaults.  The only change
which can be made to this order is to instruct the configuration file
action to add options inline rather than at the beginning, causing
configuration options to be parsed after environment variables and in
the order encountered.
"""

import argparse as _argparse
import os as _os
import re as _re
import shlex as _shlex
import sys as _sys


def expandall(string):
    return _os.path.expanduser(_os.path.expandvars(string))


class TrackingString(str):
    """String subclass that allows attaching source information."""

    def __new__(cls, *args, **kwargs):
        source = kwargs.pop("source", None)
        obj = str.__new__(cls, *args, **kwargs)
        obj.source = source
        return obj


class AddConstAction(_argparse.Action):
    """Add a constant value to the option."""

    def __init__(
        self,
        option_strings,
        dest,
        const=1,
        type=int,
        default=None,
        required=False,
        help=None,
    ):
        super(AddConstAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            type=type,
            const=const,
            default=default,
            required=required,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        value = getattr(namespace, self.dest, None) or 0
        setattr(namespace, self.dest, value + self.const)


class ListAction(_argparse.Action):
    """Action to store space or comma separated lists."""

    def split(self, value):
        buf = []
        esc = False
        for c in value:
            if esc:
                buf.append(c)
                esc = False
            elif c == "\\":
                esc = True
            elif c.isspace() or c == ",":
                if buf:
                    yield "".join(buf)
                    del buf[:]
            else:
                buf.append(c)
        if buf:
            yield "".join(buf)

    def __call__(self, parser, namespace, value, option_string=None):
        if value and value[:1] in "!-+":
            values = set(getattr(namespace, self.dest, []))
            list_values = set(self.split(value[1:]))
            if value[:1] == "+":
                values |= list_values
            else:
                values -= list_values
        elif not value:
            values = None
        else:
            values = set(self.split(value))
        setattr(namespace, self.dest, list(values))


class ConfigFileAction(_argparse.Action):
    """Read a configuration file and add it to the arguments to be parsed.

    The file should contain one "long option" per line, with or without
    the leading option characters (usually --), and otherwise entered as
    on the command-line with shell-style quoting. Lines beginning with
    a hash (#) or semicolon (;) and empty lines are ignored. Hashes (#)
    may also be used to comment option lines, excluding everything from
    the comment character to the end of the line. Key-value options may
    be separated by =, : or whitespace.

    By default, INI-style section markers are ignored. If, however, a
    program wants only to use settings in certain sections, it may pass
    a list of sections to allow. All other sections will be ignored.
    Section names are case-sensitive. The special None section will
    include global values which appear in the file before and section is
    declared.

    Options which set their config flag to False will not be allowed in
    configuration files. A default value can be set on the parser with
    the allow_in_config option.

    If ignore_unknown is True, lines containing an unknown option key
    will be ignored. Otherwise, an error will be issued (default
    behavior).  Options read from the configuration file will normally
    be added before those for environment variables and other
    command-line arguments and in the order encountered to maintain the
    normal order of precedence.  If inline is True, the configuration
    options will be inserted in place of the configuration option and
    will be processed after environment variables.
    """

    def __init__(self, option_strings, dest, required=False, help=None, metavar=None, **kwargs):
        ignore_unknown = kwargs.pop("ignore_unknown", None)
        inline = kwargs.pop("inline", False)
        sections = kwargs.pop("sections", None)
        super(ConfigFileAction, self).__init__(
            option_strings=option_strings,
            dest=_argparse.SUPPRESS,
            required=required,
            help=help,
            metavar=metavar,
        )
        self.ignore_unknown = ignore_unknown
        self.inline = inline
        self.sections = sections
        self.preprocess = True

    def __call__(self, parser, namespace, path, option_string=None, seen_files=None):
        # Get default for including argument in config.
        allow_in_config = getattr(parser, "allow_in_config", True)
        if seen_files is None:
            seen_files = set()
        arg_strings = []
        try:
            with open(path) as file:
                stat = _os.fstat(file.fileno())
                file_id = stat.st_dev, stat.st_ino
                # Detect config file loops
                if file_id in seen_files:
                    # XXX: Should a warning be issued?
                    return [], []
                seen_files.add(file_id)
                for section, key, args, lineno in self.itersettings(parser, file):
                    if self.sections is not None and section not in self.sections:
                        continue
                    source = ("config", (path, lineno, key))
                    opt = TrackingString((key if key.startswith("--") else ("--" + key)),
                                         source=source)
                    try:
                        action = parser._option_string_actions[opt]
                    except KeyError:
                        if self.ignore_unknown:
                            continue
                        parser.error("{} (line {}): unknown option: {}".format(path, lineno, key))
                    if not getattr(action, "config", allow_in_config):
                        parser.error("{} (line {}): option not allowed: {}".format(
                            path, lineno, key))
                    if isinstance(action, ConfigFileAction):
                        if args is None or len(args) != 1:
                            parser.error("{} (line {}): option expects one argument: {}".format(
                                path, lineno, key))
                        for arg_list in self(parser,
                                             namespace,
                                             args[0],
                                             opt,
                                             seen_files=seen_files):
                            arg_strings.extend(arg_list)
                    else:
                        if action.nargs == 0 and args:
                            if len(args) > 1:
                                parser.error("{} (line {}): option expects no "
                                             "more than one argument: {}".format(
                                                 path, lineno, key))
                            opt = parser.get_switch(action, args[0], opt)
                            arg_strings.append(TrackingString(opt, source=source))
                        else:
                            arg_strings.append(opt)
                            if args:
                                arg_strings.extend(args)
        except IOError as e:
            parser.error("{}: {}".format(path, e))
        return ([], arg_strings) if self.inline else (arg_strings, [])

    def itersettings(self, parser, conffile):
        section_re = _re.compile(r"^\s*\[\s*((?:\\.|[^\]])*?)\s*\](.*)$")
        comment_re = _re.compile(r"^\s*(?:[#;].*)?$")
        setting_re = _re.compile(r"^(\S+?)(?:(?:\s*[:=]\s*|\s+)(.*))?$")
        section = None
        for lineno, line in enumerate(conffile):
            if not line:
                break
            line = line.strip()
            if not line or comment_re.match(line):
                continue
            match = section_re.match(line)
            if match:
                section, rest = match.groups()
                if not comment_re.match(rest):
                    err = "invalid syntax after section: {!r}".format(rest)
                    parser.error("{}: {} (line {})".format(conffile.name, err, lineno))
                # Remove backslash escapes
                section = _re.sub(r"\\(.)", r"\1", section)
                continue
            key, value = setting_re.match(line).groups()
            if value is not None:
                try:
                    value = _shlex.split(value, True, True)
                except ValueError as e:
                    parser.error("{}: {} (line {})".format(conffile.name, e, lineno))
            yield section, key, value, lineno


def CaseInsensitiveConfigFileAction(ConfigFileAction):

    def itersettings(self, parser, conffile):
        itersettings = super(CaseInsensitiveConfigFileAction, self).itersettings
        for section, key, value, lineno in itersettings(parser, conffile):
            if section is not None:
                section = section.lower()
            yield section, key, value, lineno


# Revert _SubParsersAction due to Python revision 1a3143752db2
# (https://hg.python.org/cpython/rev/1a3143752db2) which introduced
# a backward-incompatible bug and broke parsing of global options.
# See Python issue #9351 (https://bugs.python.org/issue9351) for the bug
# report which spurred the change and issue 24251
# (https://bugs.python.org/issue24251) for the issue reporting the
# unintended side effect. Then override its registration below.


class SubParsersAction(_argparse._SubParsersAction):

    def __call__(self, parser, namespace, values, option_string=None):
        parser_name = values[0]
        arg_strings = values[1:]

        # set the parser name if requested
        if self.dest is not _argparse.SUPPRESS:
            setattr(namespace, self.dest, parser_name)

        # select the parser
        try:
            parser = self._name_parser_map[parser_name]
        except KeyError:
            tup = parser_name, ", ".join(self._name_parser_map)
            msg = _("unknown parser %r (choices: %s)") % tup
            raise _argparse.ArgumentError(self, msg)

        # parse all the remaining options into the namespace
        # store any unrecognized options on the object, so that the top
        # level parser can decide what to do with them
        namespace, arg_strings = parser.parse_known_args(arg_strings, namespace)

        if arg_strings:
            vars(namespace).setdefault(_argparse._UNRECOGNIZED_ARGS_ATTR, [])
            getattr(namespace, _argparse._UNRECOGNIZED_ARGS_ATTR).extend(arg_strings)


def env_var_formatter(formatter_class=_argparse.HelpFormatter):
    """Decorator to automatically add env_var documentation to help."""

    class EnvHelpFormatter(formatter_class):

        def _get_help_string(self, action):
            # pylint: disable=super-on-old-class
            help = super(EnvHelpFormatter, self)._get_help_string(action)
            if "%(env_var)" not in help:
                env_var = getattr(action, "env_var", None)
                if env_var is not None:
                    help += " (env var: %(env_var)s)"
            return help

    return EnvHelpFormatter


class ArgumentParser(_argparse.ArgumentParser):
    _false_re = _re.compile(r"^\s*(f(?:alse)?|n(?:o)?|0)?\s*$", _re.I)

    def __init__(self, *args, **kwargs):
        allow_in_config = kwargs.pop("allow_in_config", True)
        super(ArgumentParser, self).__init__(*args, **kwargs)
        self.allow_in_config = allow_in_config
        self.register("action", "add_const", AddConstAction)
        self.register("action", "store_list", ListAction)
        self.register("action", "parse_config", ConfigFileAction)
        self.register("action", "parsers", SubParsersAction)

    def _parse_known_args(self, arg_strings, namespace):
        # replace arg strings that are file references
        if self.fromfile_prefix_chars is not None:
            arg_strings = self._read_args_from_files(arg_strings)
        arg_strings = self._preprocess_args(arg_strings, namespace)
        return super(ArgumentParser, self)._parse_known_args(arg_strings, namespace)

    def _preprocess_args(self, arg_strings, namespace):
        """Pre-process arguments.

        May be overridden to change the order options are applied.
        """
        config_args, cli_args = self._parse_early_args(arg_strings, namespace)
        env_args = self._parse_environment(namespace)
        return config_args + env_args + cli_args

    def _parse_early_args(self, arg_strings, namespace):
        """Pre-parse args to expand preprocessed options.

        This may include reading options from configuration files.
        """
        config_args = []
        cli_args = []
        if not arg_strings:
            return config_args, cli_args
        if self._subparsers is not None:
            subcommands = {
                name
                for action in self._subparsers._group_actions
                if hasattr(action, "_name_parser_map") for name in action._name_parser_map
            }
        else:
            subcommands = set()
        args_enumerator = enumerate(arg_strings)
        args_iterator = (arg for i, arg in args_enumerator)

        def _take(n):
            if n > 0:
                for arg in args_iterator:
                    yield arg
                    n -= 1
                    if not n:
                        break

        take = lambda n: list(_take(n))
        for i, arg_string in args_enumerator:
            if arg_string == "--" or arg_string in subcommands:
                # All remaining arguments are considered positional
                cli_args.append(arg_string)
                break
            option_tuple = self._parse_optional(arg_string)
            if option_tuple is None or option_tuple[0] is None:
                # This argument is positional; skip further processing
                cli_args.append(arg_string)
                continue
            # Some kind of option was encountered, so deal with it
            # TODO: Argparse in python <=3.11.8 or <=3.12.2 returned a 3-tuple.
            #  Later versions returned a 4-tuple until somewhere in 3.14, after which it returns a list of 4-tuples.
            #  This code won't cover every possible case, but further investigation is required.
            #  Can we NOT subclass argparse now? It may not be necessary anymore....
            if type(option_tuple) is list and type(option_tuple[0]) is tuple:
                option_tuple = option_tuple[0]
            if type(option_tuple) is tuple:
                if len(option_tuple) == 4:
                    action, option_string, sep, explicit_arg = option_tuple
                elif len(option_tuple) == 3:
                    action, option_string, explicit_arg = option_tuple
                else:
                    raise ValueError(f'Unexpected number of elements in option_tuple: {option_tuple}. Please report to VOLTTRON development team.')
            else:
                raise ValueError(f'Unexpected output from parsing arguments. Got {option_tuple} from argparse._parse_options().  Please report to VOLTTRON development team.')
            if explicit_arg is not None:
                args = [explicit_arg]
            elif action.nargs in [_argparse.REMAINDER, _argparse.PARSER]:
                args = list(args_iterator)
            elif action.nargs is None:
                args = take(1)
            elif isinstance(action.nargs, int):
                args = take(action.nargs)
            else:
                # Consume arguments until another option is found
                args = take(1) if action.nargs is _argparse.ONE_OR_MORE else []
                n = 0
                for n, arg in enumerate(arg_strings[i + 1:]):
                    if arg == "--" or arg in subcommands:
                        break
                    option_tuple = self._parse_optional(arg)
                    if option_tuple is not None and option_tuple[0] is not None:
                        break
                args.extend(take(n))
            args_tuple = self.preprocess_option(action, namespace, args, option_string)
            if args_tuple is not None:
                extra_config, extra_cli = args_tuple
                config_args.extend(extra_config)
                cli_args.extend(extra_cli)
            elif explicit_arg is None:
                cli_args.append(arg_string)
                cli_args.extend(args)
            else:
                cli_args.append(arg_string)
        cli_args.extend(args_iterator)
        return config_args, cli_args

    def _parse_environment(self, namespace):
        """Create arguments for actions with associated environment vars."""
        arg_strings = []
        for action in self._actions:
            # Ignore positionals
            if not action.option_strings:
                continue
            try:
                value = _os.environ[action.env_var]
            except (AttributeError, KeyError, TypeError):
                continue
            for opt in action.option_strings:
                if opt.startswith("--"):
                    break
            source = ("environment", action.env_var)
            if action.nargs == 0:
                opt = TrackingString(self.get_switch(action, value, opt), source=source)
                arg_strings.append(opt)
            else:
                opt = TrackingString(opt, source=source)
                arg_strings.extend([opt, value])
        return arg_strings

    def preprocess_option(self, action, namespace, arg_strings, option_string):
        """Pre-process an action.

        If an action has a preprocess attribute which evaluates to True,
        then execute the action now. This type of action is expected to
        return None if it is to remain in the argument list unchanged.
        Otherwise, it should return two lists: the first will be
        appended to the configuration arguments and the second will
        replace the processed arguments.
        """
        if not getattr(action, "preprocess", False):
            return
        values = self._get_values(action, arg_strings)
        if action.nargs in [None, _argparse.OPTIONAL] and len(arg_strings) != 1:
            return
        return action(self, namespace, values, option_string)

    def get_switch(self, action, value, option_string):
        """Convert argument-less options when they have an argument.

        Useful for allowing options in a configuration file and
        environment variables to be set or unset based on an argument
        which looks like a boolean. To be invert-able, the action must
        set its inverse attribute to an option that will invert the
        meaning.
        """
        if self._false_re.match(value) is None:
            return option_string
        inverse = getattr(action, "inverse", None)
        if not inverse:
            self.error("option cannot be inverted: {}".format(option_string))
        return inverse

    def add_help_argument(self, *args, **kwargs):
        if not args:
            prefix_chars = self.prefix_chars
            default_prefix = "-" if "-" in prefix_chars else prefix_chars[0]
            args = [default_prefix + "h", default_prefix * 2 + "help"]
        kwargs.setdefault("action", "help")
        kwargs.setdefault("default", _argparse.SUPPRESS)
        kwargs.setdefault("help", _argparse._("show this help message and exit"))
        self.add_argument(*args, **kwargs)

    def add_version_argument(self, *args, **kwargs):
        if not args:
            prefix_chars = self.prefix_chars
            default_prefix = "-" if "-" in prefix_chars else prefix_chars[0]
            args = [default_prefix * 2 + "version"]
        kwargs.setdefault("action", "version")
        kwargs.setdefault("default", _argparse.SUPPRESS)
        kwargs.setdefault("help", _argparse._("show program's version number and exit"))
        self.add_argument(*args, **kwargs)


class TrackingArgumentParser(ArgumentParser):
    """Wrap up calls to actions to pre- and post-handlers.

    This class is useful as a base class for debugging parsers.
    """

    def _parse_known_args(self, arg_strings, namespace):
        self._setup_tracking()
        return super(TrackingArgumentParser, self)._parse_known_args(arg_strings, namespace)

    def _setup_tracking(self):

        def __call__(action, parser, namespace, values, option_string=None, **kwargs):
            source = getattr(option_string, "source", ("command-line", option_string))
            self.pre_action(action, parser, namespace, values, option_string, source)
            result = super(action.__class__, action).__call__(parser, namespace, values,
                                                              option_string, **kwargs)
            self.post_action(action, parser, namespace, values, option_string, source)
            return result

        for action in self._actions:
            cls = action.__class__
            if getattr(cls, "_trackable", False):
                continue
            action.__class__ = type(cls.__name__, (cls, ), {
                "__call__": __call__,
                "_trackable": True
            })

    def pre_action(self, action, parser, namespace, values, option_string, source):
        pass

    def post_action(self, action, parser, namespace, values, option_string, source):
        pass


class DebugArgumentParser(TrackingArgumentParser):
    """Write options to stderr as they are added to the namespace.

    This includes the source of the change. It does not include defaults.
    """

    def post_action(self, action, parser, namespace, values, option_string, source):
        if action.dest is _argparse.SUPPRESS:
            return
        value = getattr(namespace, action.dest, None)
        _sys.stderr.write("{} {} {!r} {!r}\n".format(option_string, action.dest, source, value))


def _patch_argparse():
    """Patch argparse to include additional attributes on options."""
    argparse_add_argument = _argparse._ActionsContainer.add_argument

    def add_argument(*args, **kwargs):
        config = kwargs.pop("config", True)
        env_var = kwargs.pop("env_var", None)
        inverse = kwargs.pop("inverse", None)
        action = argparse_add_argument(*args, **kwargs)
        action.config = config
        action.env_var = env_var
        action.inverse = inverse
        return action

    _argparse._ActionsContainer.add_argument = add_argument


_patch_argparse()





import argparse
import sys
import os

import gevent

from volttron.client.known_identities import AUTH


def on_message(peer, sender, bus, topic, headers, message):
    print('*'*80)
    print("Received:")
    print(f"\tpeer={peer},sender={sender},topic={topic})")
    print(f"\theaders={headers}")
    print(f"\tmessage={message}")
    print('-'*80)


def subscribe_to_bus(opts):

    if opts.identity_stage:
        print(f"Subscribing to {opts.topic}")
        from pathlib import Path
        from volttron.utils import ClientContext as cc
        from volttron.client.vip.agent import Agent as BaseAgent
        from volttron.types.auth import Credentials, VolttronCredentials
        from volttron.types.agent_context import AgentOptions
        from volttron.utils import jsonapi
        credentials_path = Path(
            cc.get_volttron_home()) / "credentials_store" / f"{opts.identity_stage}.json"
        if not credentials_path.exists():
            raise ValueError(f"Control connection credentials not found at {credentials_path}")

        credjson = jsonapi.load(credentials_path.open("r"))

        credentials = VolttronCredentials(**credjson)
        options = AgentOptions(heartbeat_autostart=False,
                                volttron_home=cc.get_volttron_home(),
                                enable_store=False)
        agent = BaseAgent(credentials=credentials, options=options, address=opts.address)
        event = gevent.event.Event()
        greenlet = gevent.spawn(agent.core.run, event)
        event.wait()

        agent.vip.pubsub.subscribe('pubsub', prefix=opts.topic, callback=on_message, all_platforms=opts.all_platforms).get(1)

        try:
            greenlet.join()
        except KeyboardInterrupt:
            #print("Complete")
            pass
        finally:
            print("Stopping Subscriptions")
            agent.core.stop()
        
        sys.exit()
               

    conn = opts.connection
    count = 0
    root_identity = f"subscriber"
    peers = sorted(conn.call("peerlist"))

    subscription_identity = f"subscriber{count}"

    while True:
        if subscription_identity in peers:
            count += 1
            subscription_identity = f"subscriber{count}"
            continue
        break
    
    # print(opts)
    # print(sys.executable)
    # print(sys.argv)
    # print(f"Creating credentials for {subscription_identity}")
    # We need to create new credentials or have one that is created already for us.
    value = conn.server.vip.rpc.call(AUTH, "create_credentials", identity=subscription_identity).get(timeout=4)
    args = sys.argv.copy()
    args.extend(['--identity-stage', subscription_identity])
    
    # This process will now execute in the currently executing process.  It is not like
    # subprocess in that it will not exit until the process ends.
    os.execvp(args[0], args=args)


    
    #opts.connection.server.vip.pubsub.publish("pubsub", topic=opts.topic, message=opts.data).get()

def add_subscribe_parser(add_parser_fn):

    subscriber = add_parser_fn("subscribe", help="Allow a subscription to listen to the bus and print out responses.")
    
    #publisher_subparser = publisher.add_subparsers(title="publish options")
    
    subscriber.add_argument("topic", help="Topic to publish to")
    subscriber.add_argument("--all-platforms", action="store_true",
                            help="Subscribe to all platforms for this topic/prefix")
    subscriber.add_argument("--identity-stage", default=None, help=argparse.SUPPRESS)
    subscriber.set_defaults(func=subscribe_to_bus)


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import os
import sys
import re
import traceback

_log = logging.getLogger(__name__)


def is_ip_private(vip_address):
    """Determines if the passed vip_address is a private ip address or not.

    :param vip_address: A valid ip address.
    :return: True if an internal ip address.
    """
    ip = vip_address.strip().lower().split("tcp://")[1]

    # https://en.wikipedia.org/wiki/Private_network

    priv_lo = re.compile(r"^127.\d{1,3}.\d{1,3}.\d{1,3}$")
    priv_24 = re.compile(r"^10.\d{1,3}.\d{1,3}.\d{1,3}$")
    priv_20 = re.compile(r"^192.168.\d{1,3}.\d{1,3}$")
    priv_16 = re.compile(r"^172.(1[6-9]|2[0-9]|3[0-1]).[0-9]{1,3}.[0-9]{1,3}$")

    return (priv_lo.match(ip) is not None or priv_24.match(ip) is not None
            or priv_20.match(ip) is not None or priv_16.match(ip) is not None)


def get_hostname():
    with open("/etc/hostname") as fp:
        hostname = fp.read().strip()

    if not hostname:
        raise ValueError("/etc/hostname file not found!")
    return hostname



import logging
import json
import time
from pathlib import Path
from typing import Dict, List, Optional, Any
import gevent
import httpx
from httpx import HTTPError

from volttron.server.decorators import service
from volttron.client.known_identities import PLATFORM_FEDERATION, PLATFORM
from volttron.client.vip.agent import Agent, Core
from volttron.server.server_options import ServerOptions
from volttron.types.auth.auth_service import AuthService
from volttron.types import MessageBus
from volttron.utils import set_agent_identity

_log = logging.getLogger(__name__)

DEFAULT_GROUP = "default"
DEFAULT_RETRY_PERIOD = 30  # seconds
logging.getLogger("httpcore.http11").setLevel(logging.WARNING)
logging.getLogger("httpcore.connection").setLevel(logging.WARNING)
class _PlatformInstance:
    """Represents a connected remote VOLTTRON platform"""
    
    def __init__(self, platform_id: str, address: str, public_credentials: str, group: str = DEFAULT_GROUP):
        self.platform_id = platform_id
        self.address = address
        self.public_credentials = public_credentials
        self.group = group
        self.connected = False
        self.last_heartbeat = None
    
    def to_dict(self) -> dict:
        """Convert to dictionary for serialization"""
        return {
            "id": self.platform_id,
            "address": self.address,
            "public_credentials": self.public_credentials,
            "group": self.group,
            "connected": self.connected,
            "last_heartbeat": self.last_heartbeat
        }
    
    @classmethod
    def from_dict(cls, data: dict) -> '_PlatformInstance':
        """Create instance from dictionary"""
        instance = cls(
            platform_id=data["id"],
            address=data["address"],
            public_credentials=data["public_credentials"],
            group=data.get("group", DEFAULT_GROUP)
        )
        instance.connected = data.get("connected", False)
        instance.last_heartbeat = data.get("last_heartbeat")
        return instance

@service
class FederationService(Agent):
    """
    Service that manages connections between multiple VOLTTRON platforms.
    Enables cross-platform message routing and RPC calls.
    """
    class Meta:
        identity = PLATFORM_FEDERATION
    
    def __init__(self, options: ServerOptions, auth_service: AuthService, messagebus: MessageBus, **kwargs):
        
        kwargs["identity"] = self.Meta.identity
        with set_agent_identity(self.Meta.identity):
            super().__init__(address=options.service_address, **kwargs)
        
        self._options = options
        self._auth_service = auth_service
        self._messagebus = messagebus
        self._connected_platforms: Dict[str, _PlatformInstance] = {}
        self._is_running = False
        self._registry_url = None
        self._httpx_client = httpx.Client(timeout=10.0)
        self._federation_config_path = Path(self._options.volttron_home) / "federation_config.json"
        self._federation_bridge = None 
        self._federation_enabled = self._options.enable_federation
        self._retry_period = DEFAULT_RETRY_PERIOD
        self._registry_connection_successful = False
        self._registry_last_attempt = 0
        
        # Load any existing federation configuration
        self._load_config()
        
        # If federation is enabled and URL is set in options, register at startup
        if options.enable_federation and options.federation_url:
            self._registry_url = options.federation_url.split('#')[0]
            
    def _load_config(self):
        """Load federation configuration from file"""
        if not self._federation_config_path.exists():
            _log.info("No federation configuration file found, creating new configuration")
            self._save_config()
            return
            
        try:
            config = json.loads(self._federation_config_path.read_text())
            self._registry_url = config.get('registry_url')
            
            # Load platforms from config
            platforms_data = config.get('platforms', [])
            for platform_data in platforms_data:
                try:
                    platform = _PlatformInstance.from_dict(platform_data)
                    self._connected_platforms[platform.platform_id] = platform
                except Exception as e:
                    _log.error(f"Error loading platform from config: {e}")
                    
            _log.info(f"Loaded federation configuration with {len(self._connected_platforms)} platforms")
        except Exception as e:
            _log.error(f"Error loading federation config: {e}")
            # Create a new config file if loading fails
            self._save_config()
    
    def _save_config(self):
        """Save federation configuration to file"""
        try:
            config = {
                'registry_url': self._registry_url,
                'platforms': [platform.to_dict() for platform in self._connected_platforms.values() if platform.platform_id != self._options.instance_name]
            }
            
            self._federation_config_path.write_text(json.dumps(config, indent=2))
            _log.debug("Saved federation configuration")
        except Exception as e:
            _log.error(f"Error saving federation config: {e}")

    def _start_federation(self):

        if self._federation_enabled:
            self._federation_bridge = self._messagebus.create_federation_bridge()
            if not self._federation_bridge:
                _log.error("Federation bridge not available, cannot start federation service")
                return
            
            # If we have a registry URL, register and discover platforms
            if self._registry_url:
                # Schedule registration with retry mechanism
                self._registry_greenlet = gevent.spawn(self._registry_connection_loop)
            else:
                # If no registry, just connect to platforms from config
                self._connect_to_stored_platforms()
                

    
    @Core.receiver("onstart")
    def _on_start(self, sender, **kwargs):
        """Handle startup tasks"""
        _log.info(f"Starting Federation Service with identity {self.core.identity}")
        self._is_running = True
        self._start_federation()        

    def _registry_connection_loop(self):
        """Periodically try to register and get platforms from the registry"""
        while self._is_running:
            try:
                if not self._registry_connection_successful:
                    # Handle registration only if not successful yet
                    self._registry_connection_successful = self._attempt_registration()
                else:
                    # If already registered, just update platform list
                    self._discover_platforms()
            except Exception as e:
                _log.error(f"Error in registry connection loop: {e}")
                self._registry_connection_successful = False
                    
            # Sleep before next attempt
            self._registry_last_attempt = time.time()
            gevent.sleep(self._retry_period)

    def _attempt_registration(self) -> bool:
        """Try to register with federation registry"""
        if not self._registry_url:
            return False
            
        _log.debug(f"Attempting to register with federation registry at {self._registry_url}")
        success = self.register_with_federation(self._registry_url)
        
        if success:
            _log.info("Successfully registered with federation registry")
            # Do an initial platform discovery
            self._discover_platforms()
        else:
            _log.warning(f"Failed to register with registry, will retry in {self._retry_period} seconds")
        
        return success

    def _discover_platforms(self):
        """Discover platforms and manage connections"""
        connected_count = self.discover_and_connect_platforms()
        _log.debug(f"Platform discovery found {connected_count} platforms")
    
    
    def _connect_to_stored_platforms(self):
        """Connect to platforms stored in configuration"""
        for platform_id, platform in list(self._connected_platforms.items()):
            try:
                _log.info(f"Connecting to stored platform {platform_id} at {platform.address}")
                self._connect_platform(platform)
            except Exception as e:
                _log.error(f"Failed to connect to stored platform {platform_id}: {e}")
            
    @Core.receiver("onstop")
    def _on_stop(self, sender, **kwargs):
        """Handle shutdown tasks"""
        _log.info(f"Stopping Federation Service")
        self._is_running = False
        
        # Stop connection loop greenlet
        if hasattr(self, '_registry_greenlet'):
            self._registry_greenlet.kill()
        
        # If registered with a federation registry, attempt to unregister
        if self._registry_url and self._options.instance_name and self._registry_connection_successful:
            try:
                self._httpx_client.delete(
                    f"{self._registry_url}/platform/{self._options.instance_name}",
                    headers={"accept": "application/json"}
                )
                _log.info(f"Unregistered from federation registry at {self._registry_url}")
            except Exception as e:
                _log.error(f"Failed to unregister from federation registry: {e}")
        
        # Close httpx client
        self._httpx_client.close()
        
        # Disconnect from all platforms
        self._disconnect_all_platforms()
        
        # Save final state to configuration
        self._save_config()
    
    def register_with_federation(self, registry_url: str) -> bool:
        """
        Register this platform with a federation registry.
        
        :param registry_url: URL of the federation registry service
        :return: True if registration was successful
        """
        self._registry_url = registry_url
        
        # Determine platform ID
        local_platform_id = self._options.instance_name
        if not local_platform_id:
            _log.error("Cannot register with federation: no platform ID available")
            return False
            
        # Get our platform's address
        if not self._options.address or len(self._options.address) == 0:
            _log.error("Cannot register with federation: no platform address available")
            return False
            
        # Use the first address as our externally reachable address
        platform_address = self._options.address[0]
        
        # Simple validation of the address format
        if not (platform_address.startswith('tcp://') or 
                platform_address.startswith('ipc://') or 
                platform_address.startswith('inproc://')):
            _log.error(f"Invalid address format for federation: {platform_address}")
            _log.error("Address should start with tcp://, ipc://, or inproc://")
            return False
        
        # Get our platform's public credentials using auth service
        public_credentials = None
        try:
            # Use auth_service directly instead of VIP
            public_credentials = self._auth_service.get_credentials(identity=PLATFORM).get_public_part()
            if not public_credentials:
                _log.error("Auth service returned empty platform public key")
        except Exception as e:
            _log.error(f"Error getting platform public key from auth service: {e}")
            
        if not public_credentials:
            _log.error("Cannot register with federation: no public credential available")
            return False
            
        # Register our platform with the federation registry
        _log.info(f"Registering platform {local_platform_id} with federation registry at {registry_url}")
        registration_data = {
            "address": platform_address,
            "group": DEFAULT_GROUP,  # Default group can be customized
            "id": local_platform_id,
            "public_credentials": public_credentials  # Match API field name
        }
        
        try:
            # Post our platform to the registry
            response = self._httpx_client.post(
                f"{registry_url}/platform", 
                headers={"accept": "application/json", "Content-Type": "application/json"},
                json=registration_data,
                timeout=5.0  # Short timeout for responsiveness
            )
            response.raise_for_status()
            _log.info(f"Successfully registered with federation registry: {response.text}")
            
            # Save the configuration with registry URL
            self._save_config()
            
            # Discover and connect to other platforms
            return self.discover_and_connect_platforms() >= 0
            
        except HTTPError as e:
            status_code = e.response.status_code if hasattr(e, 'response') else "unknown"
            _log.error(f"HTTP error registering with federation registry (status {status_code}): {e}")
            return False
        except Exception as e:
            _log.error(f"Failed to register with federation registry: {e}")
            return False
    
    def discover_and_connect_platforms(self) -> int:
        """
        Discover platforms from the registry and connect to them.
        
        :return: Number of platforms connected, -1 if error
        """
        if not self._registry_url:
            _log.error("Cannot discover platforms: no registry URL configured")
            return -1
            
        try:
            # Get list of platforms from registry
            response = self._httpx_client.get(
                f"{self._registry_url}/platforms", 
                headers={"accept": "application/json"},
                timeout=5.0  # Short timeout for responsiveness
            )
            response.raise_for_status()
            platforms_data = response.json()
            
            connected_count = 0
            local_platform_id = self._options.instance_name
            
            # Process each platform
            for platform_data in platforms_data:
                # Skip ourselves
                if platform_data["id"] == local_platform_id:
                    continue
                    
                platform_id = platform_data["id"]
                
                try:
                    # Check if we already have this platform
                    if platform_id in self._connected_platforms:
                        existing_platform = self._connected_platforms[platform_id]
                        
                        # Check if anything has changed that requires reconnection
                        if (existing_platform.address != platform_data["address"] or 
                            existing_platform.public_credentials != platform_data["public_credentials"]):
                            
                            _log.info(f"Platform {platform_id} details changed, reconnecting")
                            # Disconnect existing connection
                            try:
                                self._disconnect_platform(existing_platform)
                            except:
                                _log.warning(f"Error disconnecting from platform {platform_id} before update")
                                
                            # Create updated platform instance
                            platform = _PlatformInstance(
                                platform_id=platform_id,
                                address=platform_data["address"],
                                public_credential=platform_data["public_credentials"],
                                group=platform_data.get("group", DEFAULT_GROUP)
                            )
                            
                            # Store updated platform
                            self._connected_platforms[platform_id] = platform
                            
                            # Connect to updated platform
                            self._connect_platform(platform)
                            connected_count += 1
                        else:
                            # Platform exists with same details, just count it
                            connected_count += 1 if existing_platform.connected else 0
                    else:
                        # New platform discovered
                        _log.info(f"Discovered new platform: {platform_id}")
                        
                        # Create _PlatformInstance object
                        platform = _PlatformInstance(
                            platform_id=platform_id,
                            address=platform_data["address"],
                            public_credentials=platform_data["public_credentials"],
                            group=platform_data.get("group", DEFAULT_GROUP)
                        )
                        
                        # Store in our registry
                        self._connected_platforms[platform_id] = platform
                        
                        # Connect to the platform
                        _log.info(f"Connecting to federated platform: {platform_id} at {platform.address}")
                        self._connect_platform(platform)
                        connected_count += 1
                except Exception as e:
                    _log.error(f"Error processing platform {platform_id}: {e}")
                    
            # Save updated configuration
            self._save_config()
            
            return connected_count
            
        except HTTPError as e:
            status_code = e.response.status_code if hasattr(e, 'response') else "unknown"
            _log.error(f"HTTP error discovering platforms (status {status_code}): {e}")
            return -1
        except Exception as e:
            _log.error(f"Failed to discover platforms: {e}")
            return -1
    
    def _connect_platform(self, platform: _PlatformInstance):
        """
        Establish connection to a platform
        
        :param platform: _PlatformInstance object containing connection details
        """
        try:
            if not self._federation_bridge:
                _log.error(f"Cannot connect to platform {platform.platform_id}: federation bridge not initialized")
                return False
            
            _log.debug(f"Attempting connection to platform {platform.platform_id} at {platform.address}")
            
            # Use the existing connect method from ZmqFederationBridge
            success = self._federation_bridge.connect(
                platform_id=platform.platform_id,
                platform_address=platform.address,
                credentials=platform.public_credentials
            )
            
            if success:
                _log.info(f"Successfully established connection to platform {platform.platform_id}")
                platform.connected = True
                platform.last_heartbeat = time.time()
            else:
                _log.error(f"Failed to connect to platform {platform.platform_id}")
                platform.connected = False
                
            return success
                
        except Exception as e:
            _log.error(f"Error connecting to platform {platform.platform_id}: {e}", exc_info=True)
            platform.connected = False
            return False
    
    def _is_platform_connected(self, platform_id: str) -> bool:
        """Check if a platform is currently connected"""
        if not self._federation_bridge:
            return False
            
        try:
            # Use the federation bridge's ping method to check connection health
            return self._federation_bridge.ping(platform_id)
        except Exception as e:
            _log.error(f"Error checking connection to platform {platform_id}: {e}")
            return False
    
    def _disconnect_platform(self, platform: _PlatformInstance):
        """Disconnect from a specific platform"""
        if not self._federation_bridge:
            _log.error(f"Cannot disconnect platform {platform.platform_id}: federation bridge not initialized")
            return False
            
        try:
            # Use the federation bridge's disconnect method
            success = self._federation_bridge.disconnect(platform.platform_id)
            
            if success:
                platform.connected = False
                platform.last_heartbeat = None
                _log.info(f"Disconnected from platform {platform.platform_id}")
            else:
                _log.warning(f"Failed to disconnect from platform {platform.platform_id}")
                
            return success
        except Exception as e:
            _log.error(f"Error disconnecting from platform {platform.platform_id}: {e}", exc_info=True)
            return False
    
    def _disconnect_all_platforms(self):
        """Disconnect from all connected platforms"""
        for platform_id, platform in list(self._connected_platforms.items()):
            try:
                if platform.connected:
                    self._disconnect_platform(platform)
            except Exception as e:
                _log.error(f"Error disconnecting from platform {platform_id}: {e}")


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
import argparse


def main():
    global verbose, prompt_vhome
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("--vhome", help="Path to volttron home")
    parser.add_argument(
        "--instance-name",
        dest="instance_name",
        help="Name of this volttron instance",
    )

    group = parser.add_mutually_exclusive_group()

    agent_list = "\n\t" + "\n\t".join(sorted(available_agents.keys()))
    group.add_argument(
        "--list-agents",
        action="store_true",
        dest="list_agents",
        help="list configurable agents{}".format(agent_list),
    )

    group.add_argument("--agent", nargs="+", help="configure listed agents")
    group.add_argument(
        "--rabbitmq",
        nargs="+",
        help="Configure rabbitmq for single instance, "
        "federation, or shovel either based on "
        "configuration file in yml format or providing "
        "details when prompted. \nUsage: vcfg --rabbitmq "
        "single|federation|shovel [rabbitmq config "
        "file]",
    )
    group.add_argument(
        "--secure-agent-users",
        action="store_true",
        dest="secure_agent_users",
        help="Require that agents run with their own users (this requires running "
        "scripts/secure_user_permissions.sh as sudo)",
    )

    args = parser.parse_args()
    verbose = args.verbose
    # Protect against configuration of base logger when not the "main entry point"
    if verbose:
        setup_logging(logging.DEBUG, True)
    else:
        setup_logging(logging.INFO, True)

    prompt_vhome = True
    if args.vhome:
        set_home(args.vhome)
        prompt_vhome = False
    # if not args.rabbitmq or args.rabbitmq[0] in ["single"]:
    fail_if_instance_running()
    fail_if_not_in_src_root()
    atexit.register(_cleanup_on_exit)
    _load_config()
    if args.instance_name:
        _update_config_file(instance_name=args.instance_name)
    if args.list_agents:
        print("Agents available to configure:{}".format(agent_list))
    elif args.rabbitmq:
        if len(args.rabbitmq) > 2:
            print("vcfg --rabbitmq can at most accept 2 arguments")
            parser.print_help()
            sys.exit(1)
        elif args.rabbitmq[0] not in ["single", "federation", "shovel"]:
            print("Usage: vcf --rabbitmq single|federation|shovel "
                  "[optional path to rabbitmq config yml]")
            parser.print_help()
            sys.exit(1)
        elif len(args.rabbitmq) == 2 and not os.path.exists(args.rabbitmq[1]):
            print("Invalid rabbitmq configuration file path.")
            parser.print_help()
            sys.exit(1)
        else:
            process_rmq_inputs(args.rabbitmq, args.instance_name)
    elif args.secure_agent_users:
        config_opts["secure-agent-users"] = args.secure_agent_users
        _update_config_file()
    elif not args.agent:
        wizard()

    else:
        # Warn about unknown agents
        valid_agents = False
        for agent in args.agent:
            if agent not in available_agents:
                print('"{}" not configurable with this tool'.format(agent))
            else:
                valid_agents = True
        if valid_agents:
            confirm_volttron_home()

        # Configure agents
        for agent in args.agent:
            try:
                available_agents[agent]()
            except KeyError:
                pass



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from collections import namedtuple
import datetime
import logging
import os
import six
import time
from shutil import copyfile
from socket import gethostname, getfqdn
import subprocess

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.general_name import DNSName
from cryptography.x509.name import RelativeDistinguishedName
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID

from volttron.utils import jsonapi, ClientContext as cc, execute_command

_log = logging.getLogger(__name__)

KEY_SIZE = 1024
ENC_STANDARD = 65537
SHA_HASH = 'sha256'
# # days before the certificate will timeout.
DEFAULT_DAYS = 365 * 10
# 10 years
DEFAULT_TIMOUT = 60 * 60 * 24 * 360 * 10

PROMPT_PASSPHRASE = False


class CertError(Exception):
    pass


class Subject(
        namedtuple(
            "SubjectObj",
            ('country', 'state', 'location', 'organization', 'organization_unit', 'common_name'))):

    @staticmethod
    def create_from_x509_subject(subject):
        mapping = {
            'common_name':
            subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value,
            'country':
            subject.get_attributes_for_oid(NameOID.COUNTRY_NAME)[0].value,
            'state':
            subject.get_attributes_for_oid(NameOID.STATE_OR_PROVINCE_NAME)[0].value,
            'location':
            subject.get_attributes_for_oid(NameOID.LOCALITY_NAME)[0].value,
            'organization':
            subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)[0].value,
            'organization_unit':
            subject.get_attributes_for_oid(NameOID.ORGANIZATIONAL_UNIT_NAME)[0].value
        }
        return Subject(**mapping)


def _create_subject(**kwargs):
    """
    Create a subject to be used to create a certificate
    :param kwargs: dictionary object containing various details about who we
     are. For a self-signed certificate the subject and issuer are always the
     same.
     Possible arguments:
        C  - Country
        ST - State
        L  - Location
        O  - Organization
        OU - Organizational Unit
        CN - Common Name
    :return: Subject
    :rtype: :class:`x509.Name`
    """

    nameoid_map = {
        'C': NameOID.COUNTRY_NAME,
        'ST': NameOID.STATE_OR_PROVINCE_NAME,
        'L': NameOID.LOCALITY_NAME,
        'O': NameOID.ORGANIZATION_NAME,
        'OU': NameOID.ORGANIZATIONAL_UNIT_NAME,
        'CN': NameOID.COMMON_NAME
    }
    attributes = []
    for key in ('C', 'ST', 'L', 'O', 'OU', 'CN'):
        if key in kwargs:
            attributes.append(x509.NameAttribute(nameoid_map[key], kwargs[key]))

    subject = x509.Name(attributes)
    return subject


def _create_fingerprint(public_key):
    pub_bytes = public_key.public_bytes(serialization.Encoding.PEM,
                                        serialization.PublicFormat.SubjectPublicKeyInfo)
    h = hashes.Hash(hashes.SHA256(), default_backend())
    h.update(pub_bytes)
    return h.finalize()


def _mk_cacert(valid_days=DEFAULT_DAYS, **kwargs):
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

    issuer = subject = _create_subject(**kwargs)
    cert_builder = x509.CertificateBuilder().subject_name(subject).issuer_name(issuer).public_key(
        key.public_key()).not_valid_before(datetime.datetime.utcnow()).not_valid_after(
    # Our certificate will be valid for 365 days
            datetime.datetime.utcnow() + datetime.timedelta(days=valid_days)).add_extension(
    # set CA to true
                x509.BasicConstraints(ca=True, path_length=1),
                critical=True).serial_number(int(time.time())).add_extension(
                    x509.SubjectKeyIdentifier.from_public_key(key.public_key()), critical=False)
    cert_builder = cert_builder.add_extension(x509.KeyUsage(digital_signature=False,
                                                            key_encipherment=False,
                                                            content_commitment=False,
                                                            data_encipherment=False,
                                                            key_agreement=False,
                                                            key_cert_sign=True,
                                                            crl_sign=True,
                                                            encipher_only=False,
                                                            decipher_only=False),
                                              critical=True)
    # NOTE: M2Crypto code sets SubjectKeyIdentifier extension on
    # with cert.fingerprint() as the value. However can't do the exact same
    # thing in cryptography package as cert object is created only after I
    # call cert_builder.sign but SubjectKeyIdentifier,cert.fingerprint() can
    # be set only to certbuilder. Based on
    # https://security.stackexchange.com/questions/27797/what-damage-could-be-done-if-a-malicious-certificate-had-an-identical-subject-k
    # SubjectKeyIndentifier doesn't seem to be used in cert validation but
    # only in path generation. Though there are other documentation which
    # says it could cause problem in certificate chains. Based on reading docs
    # on the net (https://www.ietf.org/rfc/rfc3280.txt) I have created hash
    # based on public key and used that as SKI.

    cert = cert_builder.sign(key, hashes.SHA256(), default_backend())

    print("Created CA cert")
    return cert, key


def _load_cert(cert_loc):
    with open(cert_loc, 'rb') as cert_file:
        content = cert_file.read()
        cert = x509.load_pem_x509_certificate(content, default_backend())
    return cert


def get_passphrase(verify=True, prompt1='Enter passphrase:', prompt2='Verify passphrase:'):
    """
    Prompt passphrase from user and return it
    :param verify: If user should be prompt twice for verification
    :param prompt1: Prompt to be used for initial input
    :param prompt2: Prompt to used for verification
    :return: The passphrase entered by user
    :type verify: bool
    :type prompt1: str
    :type prompt2: str
    """
    from getpass import getpass
    while 1:
        try:
            p1 = getpass(prompt1)
            if verify:
                p2 = getpass(prompt2)
                if p1 == p2:
                    break
            else:
                break
        except KeyboardInterrupt:
            return None
    return p1


def _load_key(key_file_path):
    passphrase = None
    name = os.path.basename(key_file_path)
    if PROMPT_PASSPHRASE:
        passphrase = get_passphrase(prompt1="Enter passphrase for " + name,
                                    prompt2="Verify passphrase for " + name)

    with open(key_file_path, 'rb') as f:
        key = default_backend().load_pem_private_key(f.read(), passphrase)
        return key


def _get_cert_attribute_value(cert, attribute):
    try:
        return cert.subject.get_attributes_for_oid(attribute)[0].value
    except Exception as e:
        raise ValueError("Error getting value of {} :{}".format(attribute, e))


class Certs:
    """A wrapper class around certificate creation, retrieval and verification.

    """

    def remote_cert_bundle_file(self):
        return os.path.join(self.remote_cert_dir, 'requests_ca_bundle')

    def cert_file(self, name, remote=False):
        """
        Returns path to the certificate with passed name. .crt extension is
        added to the passed name be
        :param name: Name of the certificate file
        :param remote: Toggle between local and remote cert connections.
        :return: Full path the <name>.crt file
        :rtype: str
        """
        """"""
        if remote:
            return self.remote_certs_file(name)
        else:
            return '/'.join((self.cert_dir, name + '.crt'))

    def private_key_file(self, name):
        """
        return path to the private key of the passed name. Name passed should
        not contain any file extension as .pem is prefixed
        :param name: name of the key file
        :return: Full path the <name>.pem file
        :rtype: str
        """
        return '/'.join((self.private_dir, name + '.pem'))

    def ca_db_file(self, name):
        """
        return path to the ca db file of the passed name. Name passed should
        not contain any file extension
        :param name: name of the  file
        :return: Full path the <name>_cadb.json file
        :rtype: str
        """
        return '/'.join((self.ca_db_dir, name + '-cadb.json'))

    def ca_serial_file(self, name):
        """
        return the file in which ca stores the next serial number to use
        :param name: name of the ca
        :return: Full path the <name>-serial file
        :rtype: str
        """
        return '/'.join((self.ca_db_dir, name + '-serial'))

    def csr_pending_file(self, name, target=None):
        if target:
            return '/'.join((self.csr_pending_dir, target + "-" + name + ".csr"))

        return '/'.join((self.csr_pending_dir, name + ".csr"))

    def remote_certs_file(self, name):
        return '/'.join((self.remote_cert_dir, name + '.crt'))

    def __init__(self, certificate_dir=None):
        """Creates a Certs instance"""

        self.default_certs_dir = os.path.join(cc.get_volttron_home(), 'certificates')
        self.root_ca_name = cc.get_instance_name() + '-root-ca'
        self.trusted_ca_name = cc.get_instance_name() + '-trusted-cas'
        self.default_root_ca_cn = '{} {}'.format(gethostname(), self.root_ca_name)

        if not certificate_dir:
            certificate_dir = self.default_certs_dir
            # If user provided explicit directory then it should exist
            if not os.path.exists(certificate_dir):
                if certificate_dir != self.default_certs_dir:
                    raise ValueError('Invalid cert_dir {}'.format(certificate_dir))

        self.cert_dir = os.path.join(os.path.expanduser(certificate_dir), 'certs')
        self.private_dir = os.path.join(os.path.expanduser(certificate_dir), 'private')
        self.ca_db_dir = os.path.join(os.path.expanduser(certificate_dir), 'ca_db')
        self.csr_pending_dir = os.path.join(os.path.expanduser(certificate_dir), 'pending_csr')
        self.remote_cert_dir = os.path.join(os.path.expanduser(certificate_dir), 'remote_certs')
        self.certs_pending_dir = os.path.join(os.path.expanduser(certificate_dir), 'pending_certs')
        self.rejected_dir = os.path.join(os.path.expanduser(certificate_dir), 'rejected')

        required_paths = (self.cert_dir, self.private_dir, self.ca_db_dir, self.csr_pending_dir,
                          self.remote_cert_dir, self.certs_pending_dir)

        try:
            dir_created = False
            for p in required_paths:
                if not os.path.exists(p):
                    # explicitly provide rx to others since agent users should
                    # have read access to these dirs
                    os.makedirs(p)
                    os.chmod(p, 0o755)
                    dir_created = True
                else:
                    # if one exists all of them should exist. break
                    break
            if dir_created:
                os.chmod(os.path.expanduser(certificate_dir), 0o755)
        except Exception:
            raise RuntimeError("No permission to create certificates directory")

    def export_pkcs12(self, name, outfile):
        cert_file = self.cert_file(name)
        key_file = self.private_key_file(name)

        cmd = [
            "openssl", "pkcs12", "-export", "-out", outfile, "-in", cert_file, "-inkey", key_file
        ]

        subprocess.check_call(cmd)

    def ca_cert(self, public_bytes: bool = False):
        """
        Get the X509 CA certificate.
        :return: the CA certificate of current volttron instance
        """
        if not self.ca_exists():
            raise CertError("ca certificate doesn't exist")

        return self.cert(self.root_ca_name, public_bytes=public_bytes)

    def cert(self, name, remote=False, public_bytes: bool = False):
        """
        Get the X509 certificate based upon the name
        :param public_bytes:
        :param name: name of the certificate to be loaded
        :param remote: determines correct path to search for the cert.
        :return: The certificate object by the given name
        :rtype: :class: `x509._Certificate` or `byte PEM encoding`
        """

        if remote:
            cert_file = self.remote_certs_file(name)
        else:
            cert_file = self.cert_file(name)

        if not os.path.exists(cert_file):
            raise CertError("invalid certificate path {}".format(cert_file))
        cert = _load_cert(cert_file)
        if public_bytes:
            return cert.public_bytes(serialization.Encoding.PEM)

        return _load_cert(cert_file)

    def get_all_cert_subjects(self):
        subjects = []
        for fname in os.listdir(self.cert_dir):
            cert = _load_cert(self.cert_file(fname[:-4]))
            subjects.append(Subject.create_from_x509_subject(cert.subject))
        return subjects

    def get_pk_bytes(self, name):
        """
        Serialize a private key in a traditional openssl manner to be able to
        use it with JWT and other technologies.

        Traditional openssl format begins as follows

        b'-----BEGIN RSA PRIVATE KEY-----'

        :param name: full instance and identity of the key
        :return: serialized private key
        """
        pk = _load_key(self.private_key_file(name))
        return pk.private_bytes(encoding=serialization.Encoding.PEM,
                                format=serialization.PrivateFormat.TraditionalOpenSSL,
                                encryption_algorithm=serialization.NoEncryption())

    def get_cert_public_key(self, name, remote=False):
        """
        Retrieves a publickey from the passed named certificate.

        Traditional openssl format begins as follows

        b'-----BEGIN PUBLIC KEY-----'

        :param name: full instance and identity of the key
        :param remote:
        :return: serialized public key
        """
        cert = self.cert(name, remote)
        return cert.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)

    def append_external_certificate(self, certificate_str):
        return False

    def get_pending_csr_requests(self):
        pending_csr = []
        for c in os.listdir(self.csr_pending_dir):
            if c.endswith('.json'):
                if os.stat(os.path.join(self.csr_pending_dir, c)).st_size != 0:
                    with open(os.path.join(self.csr_pending_dir, c)) as fp:
                        pending_csr.append(jsonapi.loads(fp.read()))

        return pending_csr

    def get_pending_certs(self):
        return []

    def get_rejected_certs(self):
        return []

    def load_csr(self, data):
        """
        Loads a PEM X.509 CSR.
        """
        return x509.load_pem_x509_csr(data, default_backend())

    def get_csr_common_name(self, data):
        csr = self.load_csr(data)
        return csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value

    def save_pending_csr_request(self, ip_addr, common_name, csr):
        meta = dict(remote_ip_address=ip_addr,
                    identity=common_name,
                    csr=csr.decode("utf-8"),
                    status="PENDING")
        metafile = os.path.join(self.csr_pending_dir, common_name + ".json")
        csrfile = os.path.join(self.csr_pending_dir, common_name + ".csr")
        if os.path.exists(metafile):
            _log.debug("csr file already exists, not saving")
        else:
            with open(metafile, 'w') as fp:
                fp.write(jsonapi.dumps(meta))
            with open(csrfile, "wb") as fw:
                fw.write(csr)
        return csrfile

    def create_csr(self, fully_qualified_identity, remote_instance_name):
        """
        Create a csr with name as the common name.

        The key that is used to sign the csr is <_instance_name>.name.

        :param fully_qualified_identity:
        :param remote_instance_name:
        :return csr.public_bytes: Encoded certificate which can saved
        to a file or sent to be verified by clients.
        """
        assert fully_qualified_identity
        remote_rmq_user = "{}.{}".format(remote_instance_name, fully_qualified_identity)
        _, _, xname = build_subject(self.cert(self.root_ca_name), six.u(remote_rmq_user))
        key = _load_key(self.private_key_file(fully_qualified_identity))
        csr = x509.CertificateSigningRequestBuilder().subject_name(xname).sign(
            key, hashes.SHA256(), default_backend())
        # with open(self.csr_create_file(name, target_volttron), "wb") as fw:
        #     fw.write(csr.public_bytes(serialization.Encoding.PEM))
        return csr.public_bytes(serialization.Encoding.PEM)

    def get_csr_status(self, common_name):
        metafile = os.path.join(self.csr_pending_dir, common_name + ".json")
        if not os.path.isfile(metafile):
            return "UNKNOWN"

        meta = jsonapi.loads(open(metafile, 'rb').read())
        return meta.get("status")

    def get_cert_from_csr(self, common_name):
        status = self.get_csr_status(common_name)
        if status == 'APPROVED':
            return self.cert(common_name, True).public_bytes(encoding=serialization.Encoding.PEM)

    def approve_csr(self, common_name):
        metafile = os.path.join(self.csr_pending_dir, common_name + ".json")
        csrfile = os.path.join(self.csr_pending_dir, common_name + ".csr")
        if not os.path.isfile(metafile):
            raise ValueError("Unknown csr for common_name {}".format(common_name))
        if not os.path.isfile(csrfile):
            raise ValueError("Bad state unknown CSR for common_name {}".format(common_name))

        cert = self.sign_csr(csrfile)
        self.save_remote_cert(common_name, cert)
        meta = jsonapi.loads(open(metafile, 'r').read())
        meta['status'] = 'APPROVED'
        with open(metafile, 'w') as fp:
            fp.write(jsonapi.dumps(meta))
        return cert

    def delete_csr(self, common_name):
        metafile = os.path.join(self.csr_pending_dir, common_name + ".json")
        csrfile = os.path.join(self.csr_pending_dir, common_name + ".csr")

        self.delete_remote_cert(common_name)
        if os.path.exists(metafile):
            os.remove(metafile)
        if os.path.exists(csrfile):
            os.remove(csrfile)

    def deny_csr(self, common_name):
        metafile = os.path.join(self.csr_pending_dir, common_name + ".json")
        csrfile = os.path.join(self.csr_pending_dir, common_name + ".csr")

        if not os.path.isfile(metafile):
            raise ValueError("Unknown csr for common_name {}".format(common_name))
        if not os.path.isfile(csrfile):
            raise ValueError("Bad state unknown CSR for common_name {}".format(common_name))

        self.delete_remote_cert(common_name)
        meta = jsonapi.loads(open(metafile, 'r').read())
        meta['status'] = 'DENIED'

        with open(metafile, 'w') as fp:
            fp.write(jsonapi.dumps(meta))

    def sign_csr(self, csr_file):
        with open(csr_file, 'rb') as f:
            csr = x509.load_pem_x509_csr(data=f.read(), backend=default_backend())

        subject_common_name = csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
        cert, _ = self.create_signed_cert_files(name=subject_common_name, overwrite=False, csr=csr)
        return cert.public_bytes(encoding=serialization.Encoding.PEM)

    def cert_exists(self, cert_name, remote=False):
        """
        Verifies that the cert exists by filename.
        :param cert_name: name of the cert to look up
        :return: True if cert exists, False otherwise
        """
        if remote:
            return os.path.exists(self.remote_certs_file(cert_name))
        else:
            return os.path.exists(self.cert_file(cert_name))

    def ca_exists(self):
        """
        Returns true if the ca cert has been created already
        :return: True if CA cert exists, False otherwise
        """
        return os.path.exists(self.cert_file(self.root_ca_name))

    def get_cert_subject(self, name):
        """
        Retrieves the subject details of a certificate
        :param name: name of the certificate
        :return: dictionary object with the format
        {
            'country':value,
            'state': value,
            'location': value,
            'organization': value,
            'organization-unit': value,
            'common-name':value
        }
        """

        subject = self.cert(name).subject
        return {
            'common-name':
            subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value,
            'country':
            subject.get_attributes_for_oid(NameOID.COUNTRY_NAME)[0].value,
            'state':
            subject.get_attributes_for_oid(NameOID.STATE_OR_PROVINCE_NAME)[0].value,
            'location':
            subject.get_attributes_for_oid(NameOID.LOCALITY_NAME)[0].value,
            'organization':
            subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)[0].value,
            'organization-unit':
            subject.get_attributes_for_oid(NameOID.ORGANIZATIONAL_UNIT_NAME)[0].value
        }

    @staticmethod
    def get_admin_cert_names(instance_name):
        """
        Returns the name of the instance ca certificate(root ca), instance
        server certificate and instance client (admin user) certificate
        :param instance_name: name of the volttron instance
        :return: names of volttron instance certs
        """
        return instance_name + '-root-ca', instance_name + "-server", \
               instance_name + "-admin"

    @staticmethod
    def validate_key_pair(public_key_file, private_key_file):
        """
        Given a public private key pair, validate the pair.
        :param public_key_file: path to public certificate file
        :param private_key_file: path to private key file
        :return True if the pair is valid, False otherwise
        """
        try:
            cmd = [
                'openssl', 'x509', '-noout', '-modulus', '-in',
                os.path.expanduser(os.path.expandvars(public_key_file))
            ]
            mod_pub = execute_command(cmd, err_prefix="Error getting modulus of "
                                      "public key")
            cmd = [
                'openssl', 'rsa', '-noout', '-modulus', '-in',
                os.path.expanduser(os.path.expandvars(private_key_file))
            ]
            mod_key = execute_command(cmd, err_prefix="Error getting modulus of "
                                      "private key")
        except RuntimeError as e:
            return False

        return mod_pub == mod_key

    def save_agent_remote_info(self, directory, local_keyname, remote_cert_name, remote_cert,
                               remote_ca_name, remote_ca_cert):
        """
        Save the remote info file, remote certificates and remote ca to the proper place
        in the remote_certificate directory.

        :param local_keyname: identity of the local agent connected to the local messagebux
        :param remote_cert_name: identity of the dynamic agent connected to the remote message bus
        :param remote_cert: certificate returned from the remote instance
        :param remote_ca_name: name of the remote ca
        :param remote_ca_cert: certificate of the remote ca certificate
        """
        try:
            self.save_remote_cert(remote_cert_name, remote_cert, directory)
            self.save_remote_cert(remote_ca_name, remote_ca_cert, directory)
            self.create_requests_ca_bundle(directory)

            metadata = dict(remote_ca_name=remote_ca_name, local_keyname=local_keyname)
            metafile = os.path.join(directory, remote_cert_name + ".json")

            with open(metafile, 'w') as fp:
                fp.write(jsonapi.dumps(metadata))
            # Change group+other permissions to read only
            for root, dirs, files in os.walk(directory):
                for f in files:
                    os.chmod(os.path.join(root, f), 0o644)
        except Exception as e:
            _log.error(f"Error saving agent remote cert info. Exception:{e}")
            raise e

    def create_requests_ca_bundle(self, agent_remote_cert_dir):
        # if this is called by agent there will be an agent specific
        # remote cert dir in agent isolation mode
        bundle_file = os.path.join(agent_remote_cert_dir, "requests_ca_bundle")

        with open(bundle_file, 'wb') as fp:
            # First include this platforms ca
            fp.write(self.ca_cert(public_bytes=True))
            for f in os.listdir(agent_remote_cert_dir):
                # based upon the call to the save_agent_remote_info from
                # subsystem.auth file there will be a _ca added to the
                # instance name on the other side of the connection so we can
                # safely look for that string and bundle together.
                if not f.endswith("_ca.crt"):
                    continue
                filepath = os.path.join(agent_remote_cert_dir, f)

                with open(filepath, 'rb') as fr:
                    fp.write(fr.read())
        os.chmod(bundle_file, 0o664)
        _log.debug(f"Updated request ca bundle {bundle_file}")

    def delete_remote_cert(self, name):
        cert_file = self.remote_certs_file(name)
        if os.path.exists(cert_file):
            os.remove(cert_file)

    def save_remote_cert(self, name, cert_string, remote_cert_dir=None):
        if remote_cert_dir:
            # agent has its own remote cert dir in agent isolation mode
            cert_file = os.path.join(remote_cert_dir, name + ".crt")
        else:
            # default platform remote cert dir
            cert_file = self.remote_certs_file(name)
        try:
            with open(cert_file, 'wb') as fp:
                fp.write(cert_string)
        except Exception as e:
            raise RuntimeError("Error saving remote cert {}. "
                               "Exception: {}".format(cert_file, e))

    def save_cert(self, file_path):
        cert_file = self.cert_file(os.path.splitext(os.path.basename(file_path))[0])
        directory = os.path.dirname(cert_file)
        if not os.path.exists(directory):
            # make certs directory accessible to all.
            os.makedirs(directory, mode=0o755)
        if file_path != cert_file:
            copyfile(file_path, cert_file)
            # but restrict file access. even to group. umask won't change
            # group permissions
            os.chmod(cert_file, 0o644)

    def save_key(self, file_path):
        key_file = self.private_key_file(os.path.splitext(os.path.basename(file_path))[0])
        directory = os.path.dirname(key_file)
        if not os.path.exists(directory):
            # make directory accessible to all.
            os.makedirs(directory, mode=0o755)
        if file_path != key_file:
            copyfile(file_path, key_file)
            # but restrict file access. even to group. umask won't change
            # group permissions
            os.chmod(key_file, 0o600)

    def create_signed_cert_files(self,
                                 name,
                                 cert_type='client',
                                 ca_name=None,
                                 overwrite=True,
                                 valid_days=DEFAULT_DAYS,
                                 csr=None,
                                 **kwargs):
        """
        Create a new certificate and sign it with the volttron instance's
        CA certificate. Save the created certificate and the private key of
        the certificate with the given name
        :param valid_days: number of days for which cert should be valid
        :param ca_name: name of the ca to sign this cert
        :param cert_type: client or server
        :param overwrite: boolean to denote if existing cert should be
         overwritten
        :param name: name used to save the newly created certificate and
         private key. Files are saved as <name>.crt and <name>.pem
        :param csr: Certificate Signing Request(CSR) based on which cert should be created.
         In this case no new private key is generated. CSR's public bytes and subject are used in building the
         certificate
        :param kwargs: dictionary object containing various details about who we
         are.
         Possible arguments:
             C  - Country
             ST - State
             L  - Location
             O  - Organization
             OU - Organizational Unit
             CN - Common Name
        :return: True if certificate creation was successful
        """
        if csr:
            remote = True
        else:
            remote = False

        if not overwrite and self.cert_exists(name, remote=remote):
            if remote:
                return _load_cert(self.cert_file(name, remote)), None
            else:
                return _load_cert(self.cert_file(name)), self.private_key_file(name)

        if not ca_name:
            ca_name = self.root_ca_name

        cert, key, serial = _create_signed_certificate(ca_cert=self.cert(ca_name),
                                                       ca_key=_load_key(
                                                           self.private_key_file(ca_name)),
                                                       name=name,
                                                       valid_days=valid_days,
                                                       type=cert_type,
                                                       csr=csr,
                                                       **kwargs)
        if csr:
            self._save_cert(name, cert, key, remote=remote)
        else:
            self._save_cert(name, cert, key)
        self.update_ca_db(cert, ca_name, serial)
        return cert, key

    def _save_cert(self, name, cert, pk, remote=False):
        """
        Save the given certificate and private key using name.crt and
        name.pem respectively.
        :param name: File name to be used to save
        :param cert: :class: `x509._Certificate` object
        :param pk:  :class: `
        :return:
        """
        with open(self.cert_file(name, remote=remote), "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))
        os.chmod(self.cert_file(name, remote=remote), 0o644)

        if pk:
            encryption = serialization.NoEncryption()
            if PROMPT_PASSPHRASE:
                encryption = serialization.BestAvailableEncryption(
                    get_passphrase(prompt1='Enter passphrase for private '
                                   'key ' + name + ":"))

            # Write our key to disk for safe keeping
            key_file = self.private_key_file(name)
            with open(key_file, "wb") as f:
                f.write(
                    pk.private_bytes(encoding=serialization.Encoding.PEM,
                                     format=serialization.PrivateFormat.TraditionalOpenSSL,
                                     encryption_algorithm=encryption))
            os.chmod(key_file, 0o600)

    def update_ca_db(self, cert, ca_name, serial):
        """
        Update the CA db with details of the file that the ca signed.
        :param cert: cert that was signed by ca_name
        :param ca_name: name of the ca that signed the cert
        """
        db_file = self.ca_db_file(ca_name)
        ca_db = {}
        dn = "C={}/ST={}/L={}/O={}/OU={}/CN={}".format(
            _get_cert_attribute_value(cert, NameOID.COUNTRY_NAME),
            _get_cert_attribute_value(cert, NameOID.STATE_OR_PROVINCE_NAME),
            _get_cert_attribute_value(cert, NameOID.LOCALITY_NAME),
            _get_cert_attribute_value(cert, NameOID.ORGANIZATION_NAME),
            _get_cert_attribute_value(cert, NameOID.ORGANIZATIONAL_UNIT_NAME),
            _get_cert_attribute_value(cert, NameOID.COMMON_NAME))
        if os.path.exists(db_file):
            with open(db_file, "r") as f:
                ca_db = jsonapi.load(f)
        entries = ca_db.get(dn, {})
        entries['status'] = "valid"
        entries['expiry'] = cert.not_valid_after.strftime("%Y-%m-%d "
                                                          "%H:%M:%S.%f%z")
        entries['serial_number'] = cert.serial_number
        ca_db[dn] = entries
        with open(db_file, 'w+') as outfile:
            jsonapi.dump(ca_db, outfile, indent=4)

        with open(self.ca_serial_file(ca_name), "w+") as f:
            f.write(str(serial + 1))    # next available serial is current + 1

    def verify_cert(self, cert_name):
        """
        Verify a the given cert is signed by the root ca.
        :param cert_name: The name of the certificate to be verified against
        the CA
        :return:
        """
        cacert = self.ca_cert()
        cert = self.cert(cert_name)
        return cert.verify(cacert.get_pubkey())

    def create_root_ca(self, overwrite=True, valid_days=DEFAULT_DAYS, **kwargs):
        """
        Create a CA certificate with the given args and save it with the given
        name
        :param overwrite: boolan to indicate if we should overwrite
         any existing CA
        :param valid_days: Number of days for which the certificate should be
         valid. Defaults to 365 days
        :param kwargs: Details about the certificate.
         Possible arguments:
            C  - Country
            ST - State
            L  - Location
            O  - Organization
            OU - Organizational Unit
            CN - Common Name
        :return:
        """
        if not overwrite:
            if self.ca_exists():
                return

        if 'CN' not in kwargs.keys() or kwargs['CN'] is None:
            kwargs['CN'] = self.default_root_ca_cn

        cert, pk = _mk_cacert(valid_days=valid_days, **kwargs)

        self._save_cert(self.root_ca_name, cert, pk)
        return cert, pk


def _create_private_key():
    return rsa.generate_private_key(public_exponent=65537,
                                    key_size=2048,
                                    backend=default_backend())


def _create_signed_certificate(ca_cert,
                               ca_key,
                               name,
                               valid_days=DEFAULT_DAYS,
                               type='client',
                               csr=None,
                               **kwargs):
    """
    Creates signed cert of type provided and signs it with ca_key provided. To create subject for the new certificate
    common name is set new value, rest of the attributes are copied from subject of provided ca certificate
    :param ca_cert:
    :param ca_key:
    :param name:
    :param valid_days:
    :param type:
    :param kwargs:
    :return:
    """
    issuer = ca_cert.subject
    # cryptography 2.7
    ski = x509.SubjectKeyIdentifier.from_public_key(ca_cert.public_key())
    fqdn = kwargs.pop('fqdn', None)

    if csr:
        key = None
        public_key = csr.public_key()
    else:
        key = _create_private_key()
        public_key = key.public_key()

    if csr:
        subject = csr.subject
    elif kwargs:
        subject = _create_subject(**kwargs)
    else:
        fqdn, hostname, subject = build_subject(ca_cert, name, type, fqdn)

    cert_builder = x509.CertificateBuilder().subject_name(subject).issuer_name(issuer).public_key(
        public_key).not_valid_before(datetime.datetime.utcnow()).not_valid_after(
    # Our certificate will be valid for 3650 days
            datetime.datetime.utcnow() + datetime.timedelta(days=valid_days)).add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_cert.public_key()),
                critical=False)
    if type == 'CA':
        # create a intermediate CA
        cert_builder = cert_builder.add_extension(x509.BasicConstraints(ca=True, path_length=0),
                                                  critical=True).add_extension(
                                                      x509.SubjectKeyIdentifier.from_public_key(
                                                          key.public_key()),
                                                      critical=False)
        # .add_extension(
        #     x509.SubjectKeyIdentifier(
        #         _create_fingerprint(public_key)),
        #     critical=False
        # )

    else:
        # if type is server or client.
        cert_builder = cert_builder.add_extension(x509.KeyUsage(digital_signature=True,
                                                                key_encipherment=True,
                                                                content_commitment=False,
                                                                data_encipherment=False,
                                                                key_agreement=False,
                                                                key_cert_sign=False,
                                                                crl_sign=False,
                                                                encipher_only=False,
                                                                decipher_only=False),
                                                  critical=True)

    if type == 'server':
        # if server cert specify that the certificate can be used as an SSL
        # server certificate
        cert_builder = cert_builder.add_extension(x509.ExtendedKeyUsage(
            (ExtendedKeyUsageOID.SERVER_AUTH, )),
                                                  critical=False)
        if hostname and fqdn != hostname:
            cert_builder = cert_builder.add_extension(x509.SubjectAlternativeName(
                [DNSName(hostname), DNSName(fqdn)]),
                                                      critical=True)
        else:
            cert_builder = cert_builder.add_extension(x509.SubjectAlternativeName([DNSName(fqdn)]),
                                                      critical=True)

    elif type == 'client':
        # specify that the certificate can be used as an SSL
        # client certificate to enable TLS Web Client Authentication
        cert_builder = cert_builder.add_extension(x509.ExtendedKeyUsage(
            (ExtendedKeyUsageOID.CLIENT_AUTH, )),
                                                  critical=False)

    # Serial must be positive integer so we are going to
    # use an increasing milliseconds serial number
    # A CA should generate unique serial numbers for each certificate it
    # generated. (signing authority + serial number) together is expected to
    #  be unique across all certificates.
    serial = int(time.time() * 10e3)
    cert_builder = cert_builder.serial_number(serial)

    # 1. version is hardcoded to 2 in Cert builder object. same as what is
    # set by old certs.py

    # 2. No way to set comment. Using M2Crypto it was set using
    # cert.add_ext(X509.new_extension('nsComment', 'SSL sever'))

    # ca_key = _load_key(self.private_key_file(ca_name))
    cert = cert_builder.sign(ca_key, hashes.SHA256(), default_backend())
    return cert, key, serial


def build_subject(ca_cert, name, type="client", fqdn=None):
    """
    Builds a x509 Name list of OID's based on a CA certificate.

    :param ca_cert: Certificate Authority used to sign the cert
    :param name: Name of the new cert
    :param type: Server or client
    :param fqdn: Fully qualified domain name
    :return fqdn: Fully qualified domain name update/pass-through
    :return hostname: Current hostname
    :return subject: x509 Name used when building a certificate
    """
    temp_list = ca_cert.subject.rdns
    new_attrs = []
    hostname = gethostname()
    for i in temp_list:
        if i.get_attributes_for_oid(NameOID.COMMON_NAME):
            if type == 'server':
                if not fqdn:
                    hostname = gethostname()
                    fqdn = getfqdn(hostname)
                new_attrs.append(
                    RelativeDistinguishedName([x509.NameAttribute(NameOID.COMMON_NAME, hostname)]))
            else:
                new_attrs.append(
                    RelativeDistinguishedName([x509.NameAttribute(NameOID.COMMON_NAME, name)]))
        else:
            new_attrs.append(i)
    subject = x509.Name(new_attrs)
    return fqdn, hostname, subject


class CertWrapper:
    """
    This class is a wrapper around the building of certificates.
    """

    @staticmethod
    def make_self_signed_ca(ca_name, **kwargs):
        """
        Creates a self signed certificate.

        :param ca_name:
        :param kwargs:
        :return:
        """
        kwargs['CN'] = ca_name
        return _mk_cacert(**kwargs)

    @staticmethod
    def make_signed_cert(ca_cert, ca_key, common_name, **kwargs):
        kwargs['CN'] = common_name
        cert, key, serial = _create_signed_certificate(ca_cert, ca_key, common_name, **kwargs)
        return cert, key

    @staticmethod
    def load_key(keyfile):
        return _load_key(keyfile)

    @staticmethod
    def load_cert(certfile):
        return _load_cert(certfile)

    @staticmethod
    def get_private_key(keyfile):
        pk = CertWrapper.load_key(keyfile)
        privatekey = pk.private_bytes(encoding=serialization.Encoding.PEM,
                                      format=serialization.PrivateFormat.TraditionalOpenSSL,
                                      encryption_algorithm=serialization.NoEncryption())
        return privatekey

    @staticmethod
    def get_cert_public_key(certfile):
        cert = CertWrapper.load_cert(certfile)
        pubkey = cert.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)

        return pubkey



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""VOLTTRON platform™ messaging utilities."""

from string import _string, Formatter

__all__ = ["normtopic", "Topic"]

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"
__copyright__ = "Copyright (c) 2016, Battelle Memorial Institute"
__license__ = "Apache 2.0"


def normtopic(topic):
    """Normalize topic, removing extra slashes and dots."""
    if not topic:
        return topic
    comps = []
    for comp in topic.split("/"):
        if comp in ("", "."):
            continue
        if comp == "..":
            comps.pop()
        else:
            comps.append(comp)
    return "/".join(comps)


class TopicFormatter(Formatter):
    """Format topic strings allowing for optional fields.

    TopicFormatter.format() works similar to the standard str.format()
    built-in, except that format strings may contain double forward
    slashes (//) to indicate break points in the string where a valid
    string may be returned if no replaceable fields occur in the
    following component.  A format field may optionally be passed
    through unsubstituted by passing a value of None for that field.

    Normal formatter processing splits the format string into a stream
    of string platformrals and replacement fields.  The replacement fields
    are replaced by values passed in via the positional and keyword
    arguments and are converted and formatted according to the field
    format before being substituted in the token stream.  Finally, the
    stream is concatenated and returned.  The first replacement field
    missing a value in the provided arguments stops processing and an
    exception is raised.  The topic formatter follows the same behavior
    except that instead of raising an exception, the previous platformral is
    truncated at the final double slash (//), and all the processed
    components are concatenated and the result returned.

    Another addition is the optional conversion specifiers. These are
    the standard conversion specifiers 's' and 'r', but upper-cased as
    'S' or 'R'.  When one of these conversion specifiers is given, the
    component will be skipped if a value is not provided for that field.

    See the Formatter documentation for the built-in string module for
    more information on formatters and the role of each method.
    """

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth, auto_arg_index=0):
        if recursion_depth < 0:
            raise ValueError("maximum string recursion exceeded")
        result = []
        for platformral, name, format_spec, conversion in self.parse(format_string):
            if conversion in ["S", "R"]:
                optional = True
                conversion = conversion.lower()
            else:
                optional = False
            if platformral:
                result.append(platformral)
            if name is None:
                continue
            try:
                obj, arg_used = self.get_field(name, args, kwargs)
            except (KeyError, AttributeError) as e:
                if platformral:
                    try:
                        platformral, _ = platformral.rsplit("//", 1)
                    except ValueError:
                        pass
                    else:
                        result[-1] = platformral
                        if optional:
                            continue
                        break
                raise e
            used_args.add(arg_used)
            if obj is None:
                obj = "{{{}{}{}{}{}}}".format(
                    name,
                    "!" if conversion else "",
                    conversion or "",
                    ":" if format_spec else "",
                    format_spec or "",
                )
            else:
                obj = self.convert_field(obj, conversion)
                format_spec, auto_arg_index = self._vformat(format_spec, args, kwargs, used_args,
                                                            recursion_depth - 1)
                obj = self.format_field(obj, format_spec)
            result.append(obj)
        return "".join(result), auto_arg_index

    def check_unused_args(self, used_args, args, kwargs):
        for name in kwargs:
            if name not in used_args:
                raise ValueError("unused keyword argument: {}".format(name))


class Topic(str):

    def __init__(self, format_string):
        """Perform minimal validation of names used in format fields."""
        for _, name, _, _ in _string.formatter_parser(format_string):
            if name is None:
                continue
            name, _ = _string.formatter_field_name_split(name)
            if isinstance(name, int) or not name:
                raise ValueError("positional format fields are not supported;"
                                 " use named format fields only")
            if name[:1].isdigit():
                raise ValueError("invalid format field name: {}".format(name))

    def __call__(self, **kwargs):
        return self.__class__(normtopic(self.vformat(kwargs)))

    def _(self, **kwargs):
        return self.__class__(self.vformat(kwargs))

    def format(self, **kwargs):
        return self.vformat(kwargs)

    def vformat(self, kwargs):
        formatter = TopicFormatter()
        return formatter.vformat(self, (), kwargs)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, super(Topic, self).__repr__())


class Header(str):
    pass



import os
import shutil
import tempfile
from pathlib import Path
import pytest
from typing import Optional, Generator, Union


@pytest.fixture
def volttron_home() -> Generator[Path, None, None]:
    """
    Create a temporary VOLTTRON_HOME directory and set it as an environment variable.
    
    This fixture creates a clean, isolated directory for each test that uses it,
    and automatically cleans it up afterward.
    
    Returns:
        Path: Path to the temporary VOLTTRON_HOME directory
    """
    # Save original environment
    original_home = os.environ.get("VOLTTRON_HOME")
    
    try:
        # Create a temporary directory
        temp_dir = tempfile.mkdtemp(prefix="volttron_home_")
        temp_path = Path(temp_dir)
        
        # Create standard subdirectories
        (temp_path / "agents").mkdir()
        (temp_path / "certificates").mkdir()
        (temp_path / "run").mkdir()
        (temp_path / "ssh").mkdir()
        (temp_path / "keystores").mkdir()
        
        # Set the environment variable - never use default
        os.environ["VOLTTRON_HOME"] = str(temp_path)
        
        # Yield the path for the test to use
        yield temp_path
        
    finally:
        # Always restore original environment, even if test fails
        if original_home is not None:
            os.environ["VOLTTRON_HOME"] = original_home
        else:
            os.environ.pop("VOLTTRON_HOME", None)
        
        # Clean up the temporary directory
        if "temp_dir" in locals():
            shutil.rmtree(temp_dir)


@pytest.fixture
def isolated_volttron_home(request) -> Generator[Path, None, None]:
    """
    Create a test-specific VOLTTRON_HOME directory with the test name encoded.
    
    This creates a persistent directory that includes the test name, useful for
    debugging tests that fail or for running specific tests with preserved state.
    
    Returns:
        Path: Path to the VOLTTRON_HOME directory
    """
    # Save original environment
    original_home = os.environ.get("VOLTTRON_HOME")
    
    try:
        # Create a test-specific directory name
        test_name = request.node.name.replace("[", "_").replace("]", "_").replace("/", "_")
        test_dir = Path(tempfile.gettempdir()) / f"volttron_home_{test_name}"
        test_dir.mkdir(exist_ok=True, parents=True)
        
        # Create standard subdirectories
        (test_dir / "agents").mkdir(exist_ok=True)
        (test_dir / "certificates").mkdir(exist_ok=True)
        (test_dir / "run").mkdir(exist_ok=True)
        (test_dir / "ssh").mkdir(exist_ok=True)
        (test_dir / "keystores").mkdir(exist_ok=True)
        
        # Set the environment variable - never use default
        os.environ["VOLTTRON_HOME"] = str(test_dir)
        
        # Yield the path for the test to use
        yield test_dir
        
    finally:
        # Always restore original environment, even if test fails
        if original_home is not None:
            os.environ["VOLTTRON_HOME"] = original_home
        else:
            os.environ.pop("VOLTTRON_HOME", None)


@pytest.fixture
def persistent_volttron_home(request) -> Generator[Path, None, None]:
    """
    Create a persistent VOLTTRON_HOME that remains after tests for debugging purposes.
    
    This fixture creates a VOLTTRON_HOME directory that is not automatically cleaned up,
    which can be helpful for debugging failing tests.
    
    Returns:
        Path: Path to the persistent VOLTTRON_HOME directory
    """
    # Save original environment
    original_home = os.environ.get("VOLTTRON_HOME")
    
    try:
        # Create a directory with the test module name
        module_name = request.module.__name__.replace(".", "_")
        persistent_dir = Path(tempfile.gettempdir()) / f"volttron_persistent_{module_name}"
        persistent_dir.mkdir(exist_ok=True, parents=True)
        
        # Create standard subdirectories
        (persistent_dir / "agents").mkdir(exist_ok=True)
        (persistent_dir / "certificates").mkdir(exist_ok=True)
        (persistent_dir / "run").mkdir(exist_ok=True)
        (persistent_dir / "ssh").mkdir(exist_ok=True)
        (persistent_dir / "keystores").mkdir(exist_ok=True)
        
        # Set the environment variable - never use default
        os.environ["VOLTTRON_HOME"] = str(persistent_dir)
        
        # Print information for debugging purposes
        print(f"\nCreated persistent VOLTTRON_HOME at: {persistent_dir}\n")
        
        # Yield the path for the test to use
        yield persistent_dir
        
    finally:
        # Always restore original environment, even if test fails
        if original_home is not None:
            os.environ["VOLTTRON_HOME"] = original_home
        else:
            os.environ.pop("VOLTTRON_HOME", None)


def create_volttron_home_with_config(config_dict: Optional[dict] = None) -> Path:
    """
    Create a VOLTTRON_HOME directory with an optional config file.
    
    Args:
        config_dict: Dictionary of configuration values to write to config file
        
    Returns:
        Path: Path to the created VOLTTRON_HOME
    """
    import json
    
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp(prefix="volttron_home_")
    temp_path = Path(temp_dir)
    
    # Create standard subdirectories
    (temp_path / "agents").mkdir()
    (temp_path / "certificates").mkdir()
    (temp_path / "run").mkdir()
    (temp_path / "ssh").mkdir()
    (temp_path / "keystores").mkdir()
    
    # Write config file if provided
    if config_dict:
        with open(temp_path / "config", "w") as f:
            json.dump(config_dict, f, indent=2)
    
    return temp_path


@pytest.fixture
def configured_volttron_home(request) -> Generator[Path, None, None]:
    """
    Create a VOLTTRON_HOME with configuration from a pytest parameter.
    
    Use with @pytest.mark.parametrize to test different configurations:
    
    @pytest.mark.parametrize("configured_volttron_home", 
                           [{"message_bus": "zmq"}, {"message_bus": "rmq"}], 
                           indirect=True)
    def test_with_different_configs(configured_volttron_home):
        ...
    
    Returns:
        Path: Path to the configured VOLTTRON_HOME
    """
    # Save original environment
    original_home = os.environ.get("VOLTTRON_HOME")
    
    try:
        # Get configuration from parameter or use empty dict
        config_dict = getattr(request, "param", {})
        
        # Create the directory with config
        temp_path = create_volttron_home_with_config(config_dict)
        
        # Set the environment variable - never use default
        os.environ["VOLTTRON_HOME"] = str(temp_path)
        
        # Yield the path for the test to use
        yield temp_path
        
    finally:
        # Always restore original environment, even if test fails
        if original_home is not None:
            os.environ["VOLTTRON_HOME"] = original_home
        else:
            os.environ.pop("VOLTTRON_HOME", None)
        
        # Clean up
        if "temp_path" in locals():
            shutil.rmtree(temp_path)

def create_volttron_home(monkeypatch) -> str:
    """
    Creates a VOLTTRON_HOME temp directory for use within a testing context.
    This function will return a string containing the VOLTTRON_HOME but will not
    set the global variable.
    :return: str: the temp directory
    """
    volttron_home = tempfile.mkdtemp(prefix="/tmp/volttron_testing").strip()
    monkeypatch.setenv("VOLTTRON_HOME", volttron_home)

    # This is needed to run tests with volttron's secure mode. Without this
    # default permissions for folders under /tmp directory doesn't not have read or execute for
    # group or others
    os.chmod(volttron_home, 0o755)

    # Move volttron_home to be one level below the mkdir so that
    # the volttron.log file is not part of the same folder for
    # observer.
    volttron_home = os.path.join(volttron_home, "volttron_home")
    os.makedirs(volttron_home)

    return volttron_home


@pytest.fixture(scope="function")
def create_volttron_home_fun_scope(monkeypatch):

    volttron_home = create_volttron_home(monkeypatch)

    yield volttron_home.strip()

    shutil.rmtree(volttron_home, ignore_errors=True)


@pytest.fixture(scope="module")
def create_volttron_home_mod_scope(monkeypatch):

    volttron_home = create_volttron_home(monkeypatch)

    yield volttron_home.strip()

    shutil.rmtree(volttron_home, ignore_errors=True)